# -*- coding: utf-8 -*-
"""
Created on Thu May 05 13:29:12 2016

@author: Suhas Somnath
"""

from __future__ import division, print_function, absolute_import
import sys
import itertools
from collections import Iterable
from multiprocessing import Pool, cpu_count
from warnings import warn
import matplotlib.pyplot as plt
import numpy as np
from time import time
from .fft import getNoiseFloor, noiseBandFilter, makeLPF, harmonicsPassFilter
from ..io.io_hdf5 import ioHDF5
from ..io.hdf_utils import getH5DsetRefs, linkRefs, getAuxData, link_as_main, copyAttributes, copy_main_attributes
from ..io.io_utils import getTimeStamp
from ..io.microdata import MicroDataGroup, MicroDataset
from ..viz.plot_utils import rainbow_plot
from ..io.translators.utils import build_ind_val_dsets

# TODO: Use filter_parms as a kwargs instead of a required input
# TODO: Phase rotation not implemented correctly. Find and use excitation frequency

###############################################################################


def test_filter(resp_wfm, filter_parms, samp_rate, show_plots=True, use_rainbow_plots=True,
                excit_wfm=None, central_resp_size=None, verbose=False):
    """
    Filters the provided response with the provided filters. Use this only to test filters.
    This function does not care about the file structure etc.
    
    Parameters
    ----------
    resp_wfm : 1D numpy float array
        Raw response waveform in the time domain
    filter_parms : dictionary
        Dictionary that contains all the filtering parameters, see Notes for details.
    samp_rate : unsigned int 
        Sampling rate in Hertz
    show_plots : (Optional) Boolean
        Whether or not to plot FFTs before and after filtering
    use_rainbow_plots : (Optional) Boolean
        Whether or not to plot loops whose color varied as a function of time
    excit_wfm : (Optional) 1D numpy float array
        Excitation waveform in the time domain. This waveform is necessary for plotting loops. 
    central_resp_size : (Optional) unsigned int
        Number of responce sample points from the center of the waveform to show in plots. Useful for SPORC
    verbose : (Optional) string
        Whether or not to print statements
    
    Returns
    -------
    filt_data : 1D numpy float array
        Filtered signal in the time domain

    Notes
    -----
    *Filter Parameters*

    noise_threshold : float
        0<1 eg 1E-4
    comb_[Hz] : Retain harmonics of frequency
        [first frequency, band width, number of harmonics]
    LPF_cutOff_[Hz] : float
        low pass frequency cut off frequency
    band_filt_[Hz] : 2D list
        [0] = center frequency, [1] = band widths
    phase_[rad] : float
        Compensation for instrumentation induced phase offset in radians
    samp_rate_[Hz] : unsigned int
        Sampling rate in Hz
    num_pix : unsigned int
        Number of pixels to filter simultaneously

    """
    num_pts = len(resp_wfm)
    
    show_loops = excit_wfm is not None and show_plots
    
    '''
    Get parameters from the dictionary.
    '''
    noise_band_filter = filter_parms.get('band_filt_[Hz]', 1)
    if isinstance(noise_band_filter, Iterable):
        noise_band_filter = noiseBandFilter(num_pts, samp_rate, noise_band_filter[0],
                                            noise_band_filter[1])
        if verbose and isinstance(noise_band_filter, Iterable):
            print('Calculated valid noise_band_filter')

    low_pass_filter = filter_parms.get('LPF_cutOff_[Hz]', -1)
    if low_pass_filter > 0:
        low_pass_filter = makeLPF(num_pts, samp_rate, low_pass_filter)
        if verbose and isinstance(low_pass_filter, Iterable):
            print('Calculated valid low pass filter')
    else:
        low_pass_filter = 1


    harmonic_filter = filter_parms.get('comb_[Hz]', 1)
    if isinstance(harmonic_filter, Iterable):
        harmonic_filter = harmonicsPassFilter(num_pts, samp_rate, harmonic_filter[0],
                                              harmonic_filter[1], harmonic_filter[2])
        if verbose and isinstance(harmonic_filter, Iterable):
            print('Calculated valid harmonic filter')

    composite_filter = noise_band_filter * low_pass_filter * harmonic_filter

    noise_floor = filter_parms.get('noise_threshold', None)
    fft_pix_data = np.fft.fftshift(np.fft.fft(resp_wfm))
    if 0 < noise_floor < 1:
        noise_floor = getNoiseFloor(fft_pix_data, noise_floor)[0]

    if show_plots:       
        l_ind = int(0.5*num_pts)
        if type(composite_filter) == np.ndarray:
            r_ind = np.max(np.where(composite_filter > 0)[0])
        else:
            r_ind = num_pts
        w_vec = np.linspace(-0.5*samp_rate, 0.5*samp_rate, num_pts)*1E-3
        if central_resp_size:
            sz = int(0.5*central_resp_size)
            l_resp_ind = -sz+l_ind
            r_resp_ind = l_ind+sz
        else:
            l_resp_ind = l_ind
            r_resp_ind = num_pts
        
        fig = plt.figure(figsize=(12, 8))
        lhs_colspan = 2
        if show_loops is False:
            lhs_colspan = 4
        else:
            ax_loops = plt.subplot2grid((2, 4), (0, 2), colspan=2, rowspan=2)
        ax_raw = plt.subplot2grid((2, 4), (0, 0), colspan=lhs_colspan)
        ax_filt = plt.subplot2grid((2, 4), (1, 0), colspan=lhs_colspan)
        axes = [ax_raw, ax_filt]
    else:
        fig = None
        axes = None

    if show_plots:
        amp = np.abs(fft_pix_data)
        ax_raw.semilogy(w_vec[l_ind:r_ind], amp[l_ind:r_ind])
        ax_raw.semilogy(w_vec[l_ind:r_ind], (composite_filter[l_ind:r_ind] + np.min(amp))*(np.max(amp)-np.min(amp)))
        if noise_floor is not None:
            ax_raw.semilogy(w_vec[l_ind:r_ind], np.ones(r_ind-l_ind)*noise_floor)
        ax_raw.set_title('Raw Signal')

    fft_pix_data *= composite_filter

    if noise_floor is not None:
        fft_pix_data[np.abs(fft_pix_data) < noise_floor] = 1E-16  # DON'T use 0 here. ipython kernel dies

    if show_plots:
        ax_filt.semilogy(w_vec[l_ind:r_ind], np.abs(fft_pix_data[l_ind:r_ind]))
        ax_filt.set_title('Filtered Signal')
        ax_filt.set_xlabel('Frequency(kHz)')
        if noise_floor is not None:
            ax_filt.set_ylim(bottom=noise_floor)  # prevents the noise threshold from messing up plots

    filt_data = np.real(np.fft.ifft(np.fft.ifftshift(fft_pix_data)))

    if show_loops:
        if use_rainbow_plots:
            rainbow_plot(ax_loops, excit_wfm[l_resp_ind:r_resp_ind], filt_data[l_resp_ind:r_resp_ind] * 1E+3)
        else:
            ax_loops.plot(excit_wfm[l_resp_ind:r_resp_ind], filt_data[l_resp_ind:r_resp_ind]*1E+3)
        ax_loops.set_title('AI vs AO') 
        ax_loops.set_xlabel('Input Bias (V)')
        ax_loops.set_ylabel('Deflection (mV)')
        axes.append(ax_loops)
        fig.tight_layout()
    return filt_data, fig, axes

# ##############################################################################


def fft_filter_dataset(h5_main, filter_parms, write_filtered=True, write_condensed=False, num_cores=None):
    # TODO: Can simplify this function substantially. Collapse / absorb the serial and parallel functions...
    """
    Filters G-mode data using specified filter parameters and writes results to file.
        
    Parameters
    ----------
    h5_main : HDF5 dataset object
        Dataset containing the raw data
    filter_parms : dictionary
        Dictionary that contains all the filtering parameters
    write_filtered : (optional) Boolean - default True
        Whether or not to write filtered data to file
    write_condensed : (optional) Boolean - default False
        Whether or not to write condensed filtered data to file
    num_cores : unsigned int
        Number of cores to use for processing data in parallel
        
    Returns
    -------
    HDF5 group reference containing filtered dataset
    """ 
    
    def __max_pixel_read(h5_raw, num_cores, store_filt=True, hot_bins=None, bytes_per_bin=2, max_RAM_gb=16):
        """
        Returns the maximum number of pixels that can be stored in memory considering the output data and the
        number of cores
        
        Parameters
        ----------
        h5_raw : HDF5 Dataset reference
            Dataset containing the raw data
        num_cores : unsigned int
            Number of cores to use for processing data in parallel
        store_filt : boolean (optional)
            Write filtered data back to h5
        hot_bins : 1D numpy array (optional)
            Bins in the frequency domain to be saved to h5
        bytes_per_bin : unsigned int (optional)
            bytes per unit in the raw data - typically 2 bytes
        max_RAM_gb : unsigned int (optional)
            Maximum system memory that can be used. This is NOT respected well. Please fix this.            
        """
        
        # double the memory requirement if storing filtered data
        if store_filt:
            bytes_per_bin *= 2   
        bytes_per_pix = h5_raw.shape[1] * bytes_per_bin
        # account for the hot bins separately
        if hot_bins is not None:
            bytes_per_pix += len(hot_bins) * 8  # complex64
        
        # Multiply this memory requirement per pixel by the number of cores
        bytes_per_pix *= num_cores
        
        max_pix = int(np.rint(max_RAM_gb*1024**3 / bytes_per_pix))
        max_pix = max(1, min(h5_raw.shape[0], max_pix))
        print('Allowed to read', max_pix, 'of', h5_raw.shape[0], 'pixels')

        return np.uint(max_pix)
        
    def __filter_chunk(raw_mat, parm_dict, recom_cores):
        """
        This function delegates the actual fitting responsibility to the
        appropriate function. Decides whether or not serial / parallel processing
        is appropriate, number of cores, number of chunks, etc.
        
        Parameters
        ----------
        raw_mat : 2D numpy array
            Raw data arranged as [repetition, points per measurement]
        parm_dict : Dictionary
            Parameters necessary for filtering
        recom_cores : unsigned int
            Number of cores to use for processing data in parallel
        
        Returns
        -------
        (noise_floors, filt_data, cond_data)
        
        noise_floors : 1D numpy array
            Contains the noise floors per set of measurements
        filt_data : 2D numpy array or None
            filtered data arranged as [repetition, points per measurement]
        cond_data : 2D complex numpy array
            [set of measurements, frequency bins containing data]

        """
        max_cores = max(1, cpu_count()-2)
            
        recom_chunks = int(raw_mat.shape[0]/recom_cores)

        print('recom cores:', recom_cores, 'Total pixels:', raw_mat.shape[0], ', Recom chunks:', recom_chunks)

        if recom_cores > 1 and recom_chunks < 10:
            min_jobs = 20
            reduced_cores = int(raw_mat.shape[0]/min_jobs)
            # intelligently set the cores now. 
            recom_cores = min(max_cores, reduced_cores)
            print('Not enough jobs per core. Reducing cores to', recom_cores)

        if recom_cores > 1:
            return filter_chunk_parallel(raw_mat, parm_dict, recom_cores)
        else:
            return filter_chunk_serial(raw_mat, parm_dict)

    max_cores = max(1, cpu_count() - 2)
    if not num_cores:
        num_cores = max_cores
    
    if write_filtered is False and write_condensed is False:
        warn('You need to write the filtered and/or the condensed dataset to the file')
        return

    if 'num_pix' not in filter_parms:
        filter_parms['num_pix'] = 1

    num_effective_pix = h5_main.shape[0]*1.0/filter_parms['num_pix']
    if num_effective_pix % 1 > 0:
        warn('Number of pixels not divisible by the number of pixels to use for FFT filter')
        return
    num_effective_pix = int(num_effective_pix)
        
    num_pts = h5_main.shape[1]*filter_parms['num_pix']

    noise_band_filter = 1
    low_pass_filter = 1
    harmonic_filter = 1

    if 'band_filt_[Hz]' in filter_parms:
        if isinstance(filter_parms['band_filt_[Hz]'], Iterable):
            band_filt = filter_parms['band_filt_[Hz]']
            noise_band_filter = noiseBandFilter(num_pts, filter_parms['samp_rate_[Hz]'], band_filt[0], band_filt[1])
    if 'LPF_cutOff_[Hz]' in filter_parms:
        if filter_parms['LPF_cutOff_[Hz]'] > 0:
            low_pass_filter = makeLPF(num_pts, filter_parms['samp_rate_[Hz]'], filter_parms['LPF_cutOff_[Hz]'])

    if 'comb_[Hz]' in filter_parms:
        if isinstance(filter_parms['comb_[Hz]'], Iterable):
            harmonic_filter = harmonicsPassFilter(num_pts, filter_parms['samp_rate_[Hz]'], filter_parms['comb_[Hz]'][0],
                                                  filter_parms['comb_[Hz]'][1], filter_parms['comb_[Hz]'][2])

    composite_filter = noise_band_filter * low_pass_filter * harmonic_filter
    
    # ioHDF now handles automatic indexing
    grp_name = h5_main.name.split('/')[-1] + '-FFT_Filtering_'

    doing_noise_floor_filter = False

    if 'noise_threshold' in filter_parms:
        if filter_parms['noise_threshold'] > 0 and filter_parms['noise_threshold'] < 1:
            ds_noise_floors = MicroDataset('Noise_Floors',
                                           data=np.zeros(shape=num_effective_pix, dtype=np.float32))
            doing_noise_floor_filter = True
        else:
            # Illegal inputs will be deleted from dictionary
            warn('Provided noise floor threshold: {} not within (0,1)'.format(filter_parms['noise_threshold']))
            del filter_parms['noise_threshold']

    if not doing_noise_floor_filter and not isinstance(composite_filter, np.ndarray):
        warn("No filtering being performed on this dataset. Exiting!")

    if isinstance(composite_filter, np.ndarray):
        ds_comp_filt = MicroDataset('Composite_Filter', np.float32(composite_filter))
                                        
    grp_filt = MicroDataGroup(grp_name, h5_main.parent.name)
    filter_parms['timestamp'] = getTimeStamp()
    filter_parms['algorithm'] = 'GmodeUtils-Parallel'
    grp_filt.attrs = filter_parms
    if isinstance(composite_filter, np.ndarray):
        grp_filt.addChildren([ds_comp_filt])
    if doing_noise_floor_filter:
        grp_filt.addChildren([ds_noise_floors])

    if write_filtered:
        ds_filt_data = MicroDataset('Filtered_Data', data=[], maxshape=h5_main.maxshape,
                                    dtype=np.float32, chunking=h5_main.chunks, compression='gzip')
        grp_filt.addChildren([ds_filt_data])
    
    hot_inds = None

    h5_pos_inds = getAuxData(h5_main, auxDataName=['Position_Indices'])[0]
    h5_pos_vals = getAuxData(h5_main, auxDataName=['Position_Values'])[0]

    if write_condensed:
        hot_inds = np.where(composite_filter > 0)[0]
        hot_inds = np.uint(hot_inds[int(0.5*len(hot_inds)):])  # only need to keep half the data
        ds_spec_inds, ds_spec_vals = build_ind_val_dsets([int(0.5*len(hot_inds))], is_spectral=True,
                                                         labels=['hot_frequencies'], units=[''], verbose=False)
        ds_spec_vals.data = np.atleast_2d(hot_inds)  # The data generated above varies linearly. Override.
        ds_cond_data = MicroDataset('Condensed_Data', data=[], maxshape=(num_effective_pix, len(hot_inds)),
                                    dtype=np.complex, chunking=(1, len(hot_inds)), compression='gzip')
        grp_filt.addChildren([ds_spec_inds, ds_spec_vals, ds_cond_data])
        if filter_parms['num_pix'] > 1:
            # need to make new position datasets by taking every n'th index / value:
            new_pos_vals = np.atleast_2d(h5_pos_vals[slice(0, None, filter_parms['num_pix']), :])
            ds_pos_inds, ds_pos_vals = build_ind_val_dsets([int(np.unique(h5_pos_inds[:, dim_ind]).size /
                                                                filter_parms['num_pix'])
                                                            for dim_ind in range(h5_pos_inds.shape[1])],
                                                           is_spectral=False,
                                                           labels=h5_pos_inds.attrs['labels'],
                                                           units=h5_pos_inds.attrs['units'], verbose=False)
            h5_pos_vals.data = np.atleast_2d(new_pos_vals)  # The data generated above varies linearly. Override.
            grp_filt.addChildren([ds_pos_inds, ds_pos_vals])

    hdf = ioHDF5(h5_main.file)
    h5_filt_refs = hdf.writeData(grp_filt)
    if isinstance(composite_filter, np.ndarray):
        h5_comp_filt = getH5DsetRefs(['Composite_Filter'], h5_filt_refs)[0]
    if doing_noise_floor_filter:
        h5_noise_floors = getH5DsetRefs(['Noise_Floors'], h5_filt_refs)[0]
    
    # Now need to link appropriately:
    if write_filtered:
        h5_filt_data = getH5DsetRefs(['Filtered_Data'], h5_filt_refs)[0]
        copyAttributes(h5_main, h5_filt_data, skip_refs=False)
        if isinstance(composite_filter, np.ndarray):
            linkRefs(h5_filt_data, [h5_comp_filt])
        if doing_noise_floor_filter:
            linkRefs(h5_filt_data, [h5_noise_floors])

        """link_as_main(h5_filt_data, h5_pos_inds, h5_pos_vals,
                     getAuxData(h5_main, auxDataName=['Spectroscopic_Indices'])[0],
                     getAuxData(h5_main, auxDataName=['Spectroscopic_Values'])[0])"""
      
    if write_condensed:
        h5_cond_data = getH5DsetRefs(['Condensed_Data'], h5_filt_refs)[0]
        if isinstance(composite_filter, np.ndarray):
            linkRefs(h5_cond_data, [h5_comp_filt])
        if doing_noise_floor_filter:
            linkRefs(h5_cond_data, [h5_noise_floors])

        if filter_parms['num_pix'] > 1:
            h5_pos_inds = getH5DsetRefs(['Position_Indices'], h5_filt_refs)[0]
            h5_pos_vals = getH5DsetRefs(['Position_Values'], h5_filt_refs)[0]

        link_as_main(h5_cond_data, h5_pos_inds, h5_pos_vals,
                     getH5DsetRefs(['Spectroscopic_Indices'], h5_filt_refs)[0],
                     getH5DsetRefs(['Spectroscopic_Values'], h5_filt_refs)[0])
        
    rot_pts = 0
    if 'phase_rot_[pts]' in filter_parms.keys():
        rot_pts = int(filter_parms['phase_rot_[pts]'])
                  
    print('Filtering data now. Be patient, this could take a few minutes') 

    max_pix = __max_pixel_read(h5_main, max(1, cpu_count() - 2), store_filt=write_filtered, hot_bins=hot_inds,
                               bytes_per_bin=2, max_RAM_gb=16)
    # Ensure that whole sets of pixels can be read.
    max_pix = np.uint(filter_parms['num_pix']*np.floor(max_pix / filter_parms['num_pix']))
    
    parm_dict = {'filter_parms': filter_parms, 'composite_filter': composite_filter,
                 'rot_pts': rot_pts, 'hot_inds': hot_inds}

    t_start = time()
    st_pix = 0
    line_count = 0
    while st_pix < h5_main.shape[0]:
        en_pix = int(min(h5_main.shape[0], st_pix + max_pix))
        print('Reading pixels:', st_pix, 'to', en_pix, 'of', h5_main.shape[0])
        raw_mat = h5_main[st_pix:en_pix, :]
        # reshape to (set of pix, data in each set of pix)
        # print 'raw mat originally of shape:', raw_mat.shape
        raw_mat = raw_mat.reshape(-1, filter_parms['num_pix'] * raw_mat.shape[1])
        num_lines = raw_mat.shape[0]
        # print 'After collapsing pixels, raw mat now of shape:', raw_mat.shape
        (nse_flrs, filt_data, cond_data) = __filter_chunk(raw_mat, parm_dict, num_cores)
        # Insert things into appropriate HDF datasets
        print('Writing filtered data to h5')
        # print 'Noise floors of shape:', nse_flrs.shape
        if doing_noise_floor_filter:
            h5_noise_floors[line_count: line_count + num_lines] = nse_flrs
        if write_condensed:
            # print('Condensed data of shape:', cond_data.shape)
            h5_cond_data[line_count: line_count + num_lines, :] = cond_data
        if write_filtered:
            # print('Filtered data of shape:', filt_data.shape)
            h5_filt_data[st_pix:en_pix, :] = filt_data
        hdf.flush()
        st_pix = en_pix

    print('FFT filtering took {} seconds'.format(time() - t_start))

    if isinstance(composite_filter, np.ndarray):
        return h5_comp_filt.parent
    if doing_noise_floor_filter:
        return h5_noise_floors.parent
              
# #############################################################################


def filter_chunk_parallel(raw_data, parm_dict, num_cores):
    # TODO: Need to check to ensure that all cores are indeed being utilized
    """
    Filters the provided dataset in parallel
    
    Parameters
    ----------
    raw_data : 2D numpy array
        Raw data arranged as [repetition, points per measurement]
    parm_dict : Dictionary
        Parameters necessary for filtering
    num_cores : unsigned int
        Number of cores to use for processing data in parallel
    
    Returns
    -------
    (noise_floors, filt_data, cond_data)
    
    noise_floors : 1D numpy array
        Contains the noise floors per set of measurements
    filt_data : 2D numpy array or None
        filtered data arranged as [repetition, points per measurement]
    cond_data : 2D complex numpy array or None
        [set of measurements, frequency bins containing data]

    """
    # Make place-holders to hold the data:
    pix_per_set = parm_dict['filter_parms']['num_pix']
    num_sets = raw_data.shape[0]
    pts_per_set = raw_data.shape[1]
    # print('sending unit filter data of size', pts_per_set)

    noise_floors = None
    noise_thresh = None
    if 'noise_threshold' in parm_dict['filter_parms']:
        noise_thresh = parm_dict['filter_parms']['noise_threshold']
        if noise_thresh > 0 and noise_thresh < 1:
            noise_floors = np.zeros(shape=num_sets, dtype=np.float32)
        else:
            noise_thresh = None

    filt_data = None
    if not parm_dict['rot_pts']:
        filt_data = np.zeros(shape=(num_sets*pix_per_set, int(pts_per_set/pix_per_set)), dtype=raw_data.dtype)

    cond_data = None
    if parm_dict['hot_inds'] is not None:
        cond_data = np.zeros(shape=(num_sets, parm_dict['hot_inds'].size), dtype=np.complex64)
        
    # Set up single parameter:
    if sys.version_info.major == 3:
        zip_fun = zip
    else:
        zip_fun = itertools.izip
    sing_parm = zip_fun(raw_data, itertools.repeat(parm_dict))
    
    # Setup parallel processing:
    # num_cores = 10
    pool = Pool(processes=num_cores, maxtasksperchild=None)
    
    # Start parallel processing:
    num_chunks = int(np.ceil(raw_data.shape[0]/num_cores))
    parallel_results = pool.imap(unit_filter, sing_parm, chunksize=num_chunks)
    pool.close()
    pool.join()
    
    print('Done parallel computing. Now extracting data and populating matrices')
    
    # Extract data for each line...
    print_set = np.linspace(0, num_sets-1, 10, dtype=int)
    for set_ind, current_results in enumerate(parallel_results):
        if set_ind in print_set:
            print('Reading...', np.rint(100 * set_ind / num_sets), '% complete')

        temp_noise, filt_data_set, cond_data_set = current_results

        if noise_thresh is not None:
            noise_floors[set_ind] = temp_noise
        if parm_dict['hot_inds'] is not None:
            cond_data[set_ind, :] = cond_data_set
        if parm_dict['rot_pts'] is not None:
            filt_data[set_ind*pix_per_set:(set_ind+1)*pix_per_set, :] = filt_data_set
            
    return noise_floors, filt_data, cond_data
 

def filter_chunk_serial(raw_data, parm_dict):
    """
    Filters the provided dataset serially
    
    Parameters
    ----------
    raw_data : 2D numpy array
        Raw data arranged as [repetition, points per measurement]
    parm_dict : Dictionary
        Parameters necessary for filtering
    
    Returns
    -------
    (noise_floors, filt_data, cond_data)
    
    noise_floors : 1D numpy array
        Contains the noise floors per set of measurements
    filt_data : 2D numpy array or None
        filtered data arranged as [repetition, points per measurement]
    cond_data : 2D complex numpy array or None
        [set of measurements, frequency bins containing data]

    """
    # Make place-holders to hold the data:
    pix_per_set = parm_dict['filter_parms']['num_pix']
    num_sets = raw_data.shape[0]
    pts_per_set = raw_data.shape[1]
    # print ('sending unit filter data of size', pts_per_set)

    noise_floors = None
    noise_thresh = None
    if 'noise_threshold' in parm_dict['filter_parms']:
        noise_thresh = parm_dict['filter_parms']['noise_threshold']
        if noise_thresh > 0 and noise_thresh < 1:
            noise_floors = np.zeros(shape=num_sets, dtype=np.float32)
        else:
            noise_thresh = None

    filt_data = None
    if parm_dict['rot_pts'] is not None:
        filt_data = np.zeros(shape=(num_sets*pix_per_set, int(pts_per_set/pix_per_set)), dtype=raw_data.dtype)

    cond_data = None
    if parm_dict['hot_inds'] is not None:
        cond_data = np.zeros(shape=(num_sets, parm_dict['hot_inds'].size), dtype=np.complex64)
    
    # Filter each line
    print_set = np.linspace(0, num_sets-1, 10, dtype=int)
    for set_ind in range(num_sets):
        if set_ind in print_set:
            print('Reading...', np.rint(100 * set_ind / num_sets), '% complete')

        # parm_dict['t_raw'] = raw_data[set_ind,:]
        (temp_noise, filt_data_set, cond_data_set) = unit_filter((raw_data[set_ind, :], parm_dict))
        if noise_thresh is not None:
            noise_floors[set_ind] = temp_noise
        if parm_dict['hot_inds'] is not None:
            cond_data[set_ind, :] = cond_data_set
        if parm_dict['rot_pts'] is not None:
            filt_data[set_ind*pix_per_set:(set_ind+1)*pix_per_set, :] = filt_data_set
            
    return noise_floors, filt_data, cond_data

     
def unit_filter(single_parm):
    """
    Filters a single instance of a signal. 
    This is the function that is called in parallel
    
    Parameters
    ----------
    single_parm : Tuple
        Parameters and data for filtering a single data instance. 
        Constructed as (raw_data_vec, parm_dict)
    
    Returns
    -------
    noise_floors : float
        Noise floor
    filt_data : 1D numpy array or None
        filtered data
    cond_data : 1D complex numpy array or None
        frequency bins containing data
    """
    # unpack all the variables from the sole input
    t_raw = single_parm[0]
    parm_dict = single_parm[1]
    # t_raw = parm_dict['t_raw']
    filter_parms = parm_dict['filter_parms']
    composite_filter = parm_dict['composite_filter']
    rot_pts = parm_dict['rot_pts']
    hot_inds = parm_dict['hot_inds']

    noise_floor = None

    t_raw = t_raw.reshape(-1)
    f_data = np.fft.fftshift(np.fft.fft(t_raw))

    if 'noise_threshold' in filter_parms:
        if filter_parms['noise_threshold'] > 0 and filter_parms['noise_threshold'] < 1:
            noise_floor = getNoiseFloor(f_data, filter_parms['noise_threshold'])[0]
            f_data[np.abs(f_data) < noise_floor] = 1E-16  # DON'T use 0 here. ipython kernel dies

    f_data = f_data * composite_filter

    cond_data = None
    filt_data = None
    if hot_inds is not None:
        cond_data = f_data[hot_inds]
    if rot_pts is not None:
        t_clean = np.real(np.fft.ifft(np.fft.ifftshift(f_data)))
        filt_mat = t_clean.reshape(filter_parms['num_pix'], -1)
        if rot_pts > 0:
            filt_data = np.roll(filt_mat, rot_pts, axis=1)
        else:
            filt_data = filt_mat
            
    return noise_floor, filt_data, cond_data

###############################################################################


def decompress_response(f_condensed_mat, num_pts, hot_inds):
    """
    Returns the time domain representation of waveform(s) that are compressed in the frequency space
    
    Parameters
    ----------
    f_condensed_mat : 1D or 2D complex numpy arrays
        Frequency domain signals arranged as [position, frequency]. 
        Only the positive frequncy bins must be in the compressed dataset. 
        The dataset is assumed to have been FFT shifted (such that 0 Hz is at the center).
    num_pts : unsigned int
        Number of points in the time domain signal
    hot_inds : 1D unsigned int numpy array
        Indices of the frequency bins in the compressed data. 
        This index array will be necessary to reverse map the condensed 
        FFT into its original form
        
    Returns
    -------
    time_resp : 2D numpy array
        Time domain response arranged as [position, time]
        
    Notes
    -----
    Memory is given higher priority here, so this function loops over the position
    instead of doing the inverse FFT on the complete data.

    """
    f_condensed_mat = np.atleast_2d(f_condensed_mat)
    hot_inds_mirror = np.flipud(num_pts - hot_inds)
    time_resp = np.zeros(shape=(f_condensed_mat.shape[0], num_pts), dtype=np.float32)
    for pos in range(f_condensed_mat.shape[0]):
        f_complete = np.zeros(shape=num_pts, dtype=np.complex)
        f_complete[hot_inds] = f_condensed_mat[pos, :]
        # Now add the mirror (FFT in negative X axis that was removed)
        f_complete[hot_inds_mirror] = np.flipud(f_condensed_mat[pos, :])
        time_resp[pos, :] = np.real(np.fft.ifft(np.fft.ifftshift(f_complete)))
    
    return np.squeeze(time_resp)


def reshape_from_lines_to_pixels(h5_main, pts_per_cycle, scan_step_x_m=1):
    """
    Breaks up the provided raw G-mode dataset into lines and pixels (from just lines)

    Parameters
    ----------
    h5_main : h5py.Dataset object
        Reference to the main dataset that contains the raw data that is only broken up by lines
    pts_per_cycle : unsigned int
        Number of points in a single pixel
    scan_step_x_m : float
        Step in meters for pixels

    Returns
    -------
    h5_resh : h5py.Dataset object
        Reference to the main dataset that contains the reshaped data
    """
    if h5_main.shape[1] % pts_per_cycle != 0:
        warn('Error in reshaping the provided dataset to pixels. Check points per pixel')
        raise ValueError
        return
    num_cols = int(h5_main.shape[1] / pts_per_cycle)

    h5_spec_vals = getAuxData(h5_main, auxDataName=['Spectroscopic_Values'])[0]
    h5_pos_vals = getAuxData(h5_main, auxDataName=['Position_Values'])[0]
    single_AO = h5_spec_vals[:, :pts_per_cycle]

    ds_spec_inds, ds_spec_vals = build_ind_val_dsets([single_AO.size], is_spectral=True,
                                                     labels=h5_spec_vals.attrs['labels'],
                                                     units=h5_spec_vals.attrs['units'], verbose=False)
    ds_spec_vals.data = np.atleast_2d(single_AO)  # The data generated above varies linearly. Override.

    ds_pos_inds, ds_pos_vals = build_ind_val_dsets([num_cols, h5_main.shape[0]], is_spectral=False,
                                                   steps=[scan_step_x_m, h5_pos_vals[1, 0]],
                                                   labels=['X', 'Y'], units=['m', 'm'], verbose=False)

    ds_reshaped_data = MicroDataset('Reshaped_Data', data=np.reshape(h5_main.value, (-1, pts_per_cycle)),
                                    compression='gzip', chunking=(10, pts_per_cycle))

    # write this to H5 as some form of filtered data.
    resh_grp = MicroDataGroup(h5_main.name.split('/')[-1] + '-Reshape_', parent=h5_main.parent.name)
    resh_grp.addChildren([ds_reshaped_data, ds_pos_inds, ds_pos_vals, ds_spec_inds, ds_spec_vals])

    hdf = ioHDF5(h5_main.file)
    print('Starting to reshape G-mode line data. Please be patient')
    h5_refs = hdf.writeData(resh_grp)

    h5_resh = getH5DsetRefs(['Reshaped_Data'], h5_refs)[0]
    # Link everything:
    linkRefs(h5_resh,
             getH5DsetRefs(['Position_Indices', 'Position_Values', 'Spectroscopic_Indices', 'Spectroscopic_Values'],
                           h5_refs))

    # Copy the two attributes that are really important but ignored:
    copy_main_attributes(h5_main, h5_resh)

    print('Finished reshaping G-mode line data to rows and columns')

    return h5_resh



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 07 16:04:34 2016

@author: Suhas Somnath
"""

from __future__ import division, print_function, absolute_import, unicode_literals
from os import path, remove  # File Path formatting
import numpy as np  # For array operations

from igor import binarywave as bw

from .translator import Translator  # Because this class extends the abstract Translator class
from .utils import generate_dummy_main_parms, build_ind_val_dsets
from ..hdf_utils import getH5DsetRefs, linkRefs
from ..io_hdf5 import ioHDF5  # Now the translator is responsible for writing the data.
from ..microdata import MicroDataGroup, \
    MicroDataset  # The building blocks for defining hierarchical storage in the H5 file


class IgorIBWTranslator(Translator):
    """
    Translates Igor Binary Wave (.ibw) files containing images or force curves to .h5
    """

    def translate(self, file_path, verbose=False):
        """
        Translates the provided file to .h5

        Parameters
        ----------
        file_path : String / unicode
            Absolute path of the .ibw file
        verbose : Boolean (Optional)
            Whether or not to show  print statements for debugging

        Returns
        -------
        h5_path : String / unicode
            Absolute path of the .h5 file
        """

        # Load the ibw file first
        ibw_obj = bw.load(file_path)
        ibw_wave = ibw_obj.get('wave')
        parm_dict = self._read_parms(ibw_wave)
        chan_labels, chan_units = self._get_chan_labels(ibw_wave)
        if verbose:
            print('Channels and units found:')
            print(chan_labels)
            print(chan_units)

        # Get the data to figure out if this is an image or a force curve
        images = ibw_wave.get('wData')
        if images.ndim == 3:  # Image stack
            if verbose:
                print('Found image stack of size {}'.format(images.shape))
            type_suffix = 'Image'

            num_rows = parm_dict['ScanLines']
            num_cols = parm_dict['ScanPoints']

            images = images.transpose(2, 0, 1)  # now ordered as [chan, Y, X] image
            images = np.reshape(images, (images.shape[0], -1, 1))  # 3D [chan, Y*X points,1]

            ds_pos_ind, ds_pos_val = build_ind_val_dsets([num_cols, num_rows], is_spectral=False,
                                                         steps=[1.0 * parm_dict['FastScanSize'] / num_cols,
                                                                1.0 * parm_dict['SlowScanSize'] / num_rows],
                                                         labels=['X', 'Y'], units=['m', 'm'], verbose=verbose)

            ds_spec_inds, ds_spec_vals = build_ind_val_dsets([1], is_spectral=True, steps=[1],
                                                             labels=['arb'], units=['a.u.'], verbose=verbose)

        else:  # single force curve
            if verbose:
                print('Found force curve of size {}'.format(images.shape))

            type_suffix = 'ForceCurve'
            images = np.atleast_3d(images)  # now [Z, chan, 1]
            images = images.transpose((1, 2, 0))  # [chan ,1, Z] force curve

            ds_pos_ind, ds_pos_val = build_ind_val_dsets([1], is_spectral=False, steps=[25E-9],
                                                         labels=['X'], units=['m'], verbose=verbose)

            ds_spec_inds, ds_spec_vals = build_ind_val_dsets([images.shape[2]], is_spectral=True, labels=['Z'],
                                                             units=['m'], verbose=verbose)
            # The data generated above varies linearly. Override.
            # For now, we'll shove the Z sensor data into the spectroscopic values.

            # Find the channel that corresponds to either Z sensor or Raw:
            try:
                chan_ind = chan_labels.index('ZSnsr')
                ds_spec_vals.data = np.atleast_2d(np.float32(images[chan_ind]))
            except ValueError:
                try:
                    chan_ind = chan_labels.index('Raw')
                    ds_spec_vals.data = np.atleast_2d(np.float32(images[chan_ind]))
                except ValueError:
                    # We don't expect to come here. If we do, spectroscopic values remains as is
                    pass

        # Prepare the list of raw_data datasets
        chan_raw_dsets = list()
        for chan_data, chan_name, chan_unit in zip(images, chan_labels, chan_units):
            ds_raw_data = MicroDataset('Raw_Data', data=np.atleast_2d(chan_data), dtype=np.float32, compression='gzip')
            ds_raw_data.attrs['quantity'] = chan_name
            ds_raw_data.attrs['units'] = [chan_unit]
            chan_raw_dsets.append(ds_raw_data)
        if verbose:
            print('Finished preparing raw datasets')

        # Prepare the tree structure
        # technically should change the date, etc.
        spm_data = MicroDataGroup('')
        global_parms = generate_dummy_main_parms()
        global_parms['data_type'] = 'IgorIBW_' + type_suffix
        global_parms['translator'] = 'IgorIBW'
        spm_data.attrs = global_parms
        meas_grp = MicroDataGroup('Measurement_000')
        meas_grp.attrs = parm_dict
        spm_data.addChildren([meas_grp])

        if verbose:
            print('Finished preparing tree trunk')

        # Prepare the .h5 file:
        folder_path, base_name = path.split(file_path)
        base_name = base_name[:-4]
        h5_path = path.join(folder_path, base_name + '.h5')
        if path.exists(h5_path):
            remove(h5_path)

        # Write head of tree to file:
        hdf = ioHDF5(h5_path)
        # spm_data.showTree()
        hdf.writeData(spm_data, print_log=verbose)

        if verbose:
            print('Finished writing tree trunk')

        # Standard list of auxiliary datasets that get linked with the raw dataset:
        aux_ds_names = ['Position_Indices', 'Position_Values', 'Spectroscopic_Indices', 'Spectroscopic_Values']

        # Create Channels, populate and then link:
        for chan_index, raw_dset in enumerate(chan_raw_dsets):
            chan_grp = MicroDataGroup('{:s}{:03d}'.format('Channel_', chan_index), '/Measurement_000/')
            chan_grp.attrs['name'] = raw_dset.attrs['quantity']
            chan_grp.addChildren([ds_pos_ind, ds_pos_val, ds_spec_inds, ds_spec_vals, raw_dset])
            h5_refs = hdf.writeData(chan_grp, print_log=verbose)
            h5_raw = getH5DsetRefs(['Raw_Data'], h5_refs)[0]
            linkRefs(h5_raw, getH5DsetRefs(aux_ds_names, h5_refs))

        if verbose:
            print('Finished writing all channels')

        hdf.close()
        return h5_path

    @staticmethod
    def _read_parms(ibw_wave):
        """
        Parses the parameters in the provided dictionary

        Parameters
        ----------
        ibw_wave : dictionary
            Wave entry in the dictionary obtained from loading the ibw file

        Returns
        -------
        parm_dict : dictionary
            Dictionary containing parameters
        """
        parm_string = ibw_wave.get('note')
        parm_string = parm_string.rstrip('\r')
        parm_list = parm_string.split('\r')
        parm_dict = dict()
        for pair_string in parm_list:
            temp = pair_string.split(':')
            if len(temp) == 2:
                temp = [item.strip() for item in temp]
                try:
                    num = float(temp[1])
                    parm_dict[temp[0]] = num
                    try:
                        if num == int(num):
                            parm_dict[temp[0]] = int(num)
                    except OverflowError:
                        pass
                except ValueError:
                    parm_dict[temp[0]] = temp[1]

        # Grab the creation and modification times:
        other_parms = ibw_wave.get('wave_header')
        for key in ['creationDate', 'modDate', 'bname']:
            try:
                parm_dict[key] = other_parms[key]
            except KeyError:
                pass
        return parm_dict

    @staticmethod
    def _get_chan_labels(ibw_wave):
        """
        Retrieves the names of the data channels and default units

        Parameters
        ----------
        ibw_wave : dictionary
            Wave entry in the dictionary obtained from loading the ibw file

        Returns
        -------
        labels : list of strings
            List of the names of the data channels
        default_units : list of strings
            List of units for the measurement in each channel
        """
        temp = ibw_wave.get('labels')
        labels = []
        for item in temp:
            if len(item) > 0:
                labels += item
        for item in labels:
            if item == '':
                labels.remove(item)

        default_units = list()
        for chan_ind, chan in enumerate(labels):
            # clean up channel names
            if chan.lower().rfind('trace') > 0:
                labels[chan_ind] = chan[:chan.lower().rfind('trace') + 5]
            # Figure out (default) units
            if chan.startswith('Phase'):
                default_units.append('deg')
            elif chan.startswith('Current'):
                default_units.append('A')
            else:
                default_units.append('m')

        return labels, default_units

    def _parse_file_path(self, input_path):
        pass

    def _read_data(self):
        pass



"""
Created on 7/17/16 10:08 AM
@author: Suhas Somnath, Chris Smith
"""

from __future__ import division, print_function, absolute_import
from warnings import warn

import numpy as np
import psutil
import scipy

from ..io.hdf_utils import checkIfMain
from ..io.io_hdf5 import ioHDF5
import multiprocessing as mp


class Process(object):
    """
    Encapsulates the typical steps performed when applying a processing function to  a dataset.

    Parameters
    ----------
    h5_main : h5py.Dataset instance
        The dataset over which the analysis will be performed. This dataset should be linked to the spectroscopic
        indices and values, and position indices and values datasets.

    """

    def __init__(self, h5_main):
        # Checking if dataset is "Main"
        if checkIfMain(h5_main):
            self.h5_main = h5_main
            self.hdf = ioHDF5(self.h5_main.file)
        else:
            warn('Provided dataset is not a "Main" dataset with necessary ancillary datasets')
            return

        # Determining the max size of the data that can be put into memory
        self._setMemoryAndCPUs()

        self._start_pos = 0
        self._end_pos = self.h5_main.shape[0]

    def _setMemoryAndCPUs(self):
        """
        Checks hardware limitations such as memory, # cpus and sets the recommended datachunk sizes and the
        number of cores to be used by analysis methods.

        Returns
        -------
        None
        
        """

        if self._parallel:
            self._maxCpus = psutil.cpu_count() - 2
        else:
            self._maxCpus = 1
        self._maxMemoryMB = psutil.virtual_memory().available / 1e6 # in MB

        self._maxDataChunk = self._maxMemoryMB / self._maxCpus

        # Now calculate the number of positions that can be stored in memory in one go.
        mb_per_position = self.h5_main.dtype.itemsize * self.h5_main.shape[1]/1e6
        self._max_pos_per_read = int(np.floor(self._maxDataChunk / mb_per_position))
        print('Allowed to read {} pixels per chunk'.format(self._max_pos_per_read))

    def _get_data_chunk(self):
        """
        Returns a chunk of data for the guess or the fit

        Parameters
        ----------
        None

        Returns
        -------
        """
        if self._start_pos < self.h5_main.shape[0]:
            self._end_pos = int(min(self.h5_main.shape[0], self._start_pos + self._max_pos_per_read))
            self.data = self.h5_main[self._start_pos:self._end_pos, :]
            print('Reading pixels {} to {} of {}'.format(self._start_pos, self._end_pos, self.h5_main.shape[0]))
        else:
            print('Finished reading all data!')
            self.data = None

    def _set_results(self):
        """
        Writes the provided guess or fit results into appropriate datasets.
        Given that the guess and fit datasets are relatively small, we should be able to hold them in memory just fine

        """
        # Now update the start position
        self._start_pos = self._end_pos
        warn('Please override the _set_results specific to your model')
        pass

    def _create_results_datasets(self):
        """
        Model specific call that will write the h5 group, guess dataset, corresponding spectroscopic datasets and also
        link the guess dataset to the spectroscopic datasets. It is recommended that the ancillary datasets be populated
        within this function.
        """
        warn('Please override the _create_results_datasets specific to your model')
        pass

    def compute(self, func, **kwargs):
        """

        Parameters
        ----------
        func
        strategy: string
            Default is 'Wavelet_Peaks'.
            Can be one of ['wavelet_peaks', 'relative_maximum', 'gaussian_processes']. For updated list, run GuessMethods.methods
        options: dict
            Default {"peaks_widths": np.array([10,200])}}.
            Dictionary of options passed to strategy. For more info see GuessMethods documentation.

        kwargs:
            processors: int
                number of processors to use. Default all processors on the system except for 1.

        Returns
        -------

        """

        self._create_results_datasets()
        self._start_pos = 0

        processors = kwargs.get("processors", self._maxCpus)
        results = list()
        if self._parallel:
            # start pool of workers
            print('Computing in parallel ... launching %i kernels...' % processors)
            pool = mp.Pool(processors)
            self._get_data_chunk()
            while self.data is not None:  # as long as we have not reached the end of this data set:
                # apply guess to this data chunk:
                tasks = [vector for vector in self.data]
                chunk = int(self.data.shape[0] / processors)
                jobs = pool.imap(func, tasks, chunksize=chunk)
                # get Results from different processes
                print('Extracting Results...')
                temp = [j for j in jobs]
                # Reformat the data to the appropriate type and or do additional computation now

                # Write to file
                self._set_results(is_guess=True)
                # read the next chunk
                self._get_data_chunk()

            # Finished reading the entire data set
            print('closing %i kernels...' % processors)
            pool.close()
        else:
            print("Computing Guesses In Serial ...")
            self._get_data_chunk()
            while self.data is not None:  # as long as we have not reached the end of this data set:
                temp = [func(vector) for vector in self.data]
                results.append(self._reformatResults(temp))
                # read the next chunk
                self._get_data_chunk()

        # reorder to get one numpy array out
        self.guess = np.hstack(tuple(results))
        print('Completed computing guess. Writing to file.')




# -*- coding: utf-8 -*-
"""
@author: Ondrej Dyck
"""

from __future__ import division, print_function, absolute_import, unicode_literals
import os
import numpy as np
from scipy.optimize import least_squares
import itertools as itt
import multiprocessing as mp
import time as tm
from _warnings import warn
from sklearn.neighbors import KNeighborsClassifier
import h5py

import matplotlib.pyplot as plt
import matplotlib.patches as patches
import sys

from ...io.io_utils import recommendCores, realToCompound
from ...io.microdata import MicroDataset, MicroDataGroup
from ...io.io_hdf5 import ioHDF5
from ...viz import plot_utils
from ..model import Model

def do_fit(single_parm):
    parms = single_parm[0]
    coef_guess_mat = parms[1]
    fit_region = parms[2]
    s1 = parms[3]
    s2 = parms[4]
    lb_mat = parms[5]
    ub_mat = parms[6]
    kwargs = single_parm[1]
    max_function_evals = kwargs['max_function_evals']
    plsq = least_squares(gauss_2d_residuals,
                         coef_guess_mat.ravel(),
                         args=(fit_region.ravel(), s1.T, s2.T),
                         kwargs=kwargs,
                         bounds=(lb_mat.ravel(), ub_mat.ravel()),
                         jac='2-point', max_nfev=max_function_evals)

    coef_fit_mat = np.reshape(plsq.x, (-1, 7))

    #if verbose:
    #    return coef_guess_mat, lb_mat, ub_mat, coef_fit_mat, fit_region, s_mat, plsq
    #else:
    return coef_fit_mat


def gauss_2d_residuals(parms_vec, orig_data_mat, x_data, y_data, **kwargs):
    """
    Calculates the residual

    Parameters
    ----------
    parms_vec : 1D numpy.ndarray
        Raveled version of the parameters matrix
    orig_data_mat : 2D numpy array
        Section of the image being fitted
    x_data : 3D numpy.ndarray

    y_data : numpy.ndarray

    Returns
    -------
    err_vec : 1D numpy.ndarray
        Difference between the original data and the matrix obtained by evaluating parms_vec with x_data_mat

    """
    # Only need to reshape the parms from 1D to 2D
    parms_mat = np.reshape(parms_vec, (-1, 7))
    # print(parms_mat)

    err = orig_data_mat - gauss2d(x_data, y_data, *parms_mat, **kwargs).ravel()

    return err


def gauss2d(X, Y, *parms, **kwargs):
    """
    Calculates a general 2d elliptic gaussian

    Parameters
    ----------

    X, Y : the x and y matrix values from the call "X, Y = np.meshgrid(x,y)" where x and y are
        defined by x = np.arange(-width/2,width/2) and y = np.arange(-height/2,height/2). 

    params: List of 7 parameters defining the gaussian.
        The parameters are [A, x0, y0, sigma_x, sigma_y, theta, background]
            A : amplitude
            x0: x position
            y0: y position
            sigma_x: standard deviation in x
            sigma_y: standard deviation in y
            theta: rotation angle
            background: a background constant

    Returns
    -------

    Returns a width x height matrix of values representing the call to the gaussian function at each position. 
    """
    symmetric = kwargs['symmetric']
    background = kwargs['background']
    Z = np.zeros(np.shape(X))
    background_value = parms[0][-1]  # we can only have one background value for the fit region
    for guess in parms:
        # each gaussian has a background associated with it but we only use the center atom background
        A, x0, y0, sigma_x, sigma_y, theta, background_unused = guess

        # determine which type of gaussian we want
        if symmetric:
            sigma_y = sigma_x
            if not background:
                background = 0
        else:
            if not background:
                background = 0

        # define some variables
        a = np.cos(theta) ** 2 / (2 * sigma_x ** 2) + np.sin(theta) ** 2 / (2 * sigma_y ** 2)
        b = -np.sin(2 * theta) / (4 * sigma_x ** 2) + np.sin(2 * theta) / (4 * sigma_y ** 2)
        c = np.sin(theta) ** 2 / (2 * sigma_x ** 2) + np.cos(theta) ** 2 / (2 * sigma_y ** 2)

        # calculate the final value
        Z += A * np.exp(- (a * (X - x0) ** 2 - 2 * b * (X - x0) * (Y - y0) + c * (Y - y0) ** 2)) + background_value
    return Z


class Gauss_Fit(object):
    """
    Initializes the gaussian fitting routines:
        fit_motif()
        fit_atom_positions_parallel()
        write_to_disk()


    Parameters
    ----------

    atom_grp : h5py.Group reference
        Parent group containing the atom guess positions, cropped clean image and motif positions

    fitting_parms : dictionary
        Parameters used for atom position fitting
        'fit_region_size': region to consider when fitting. Should be large enough to see the nearest neighbors.
        'num_nearest_neighbors': the number of nearest neighbors to fit
        'sigma_guess': starting guess for gaussian standard deviation. Should be about the size of an atom width in pixels.
        'position_range': range that the fitted position can move from initial guess position in pixels
        'max_function_evals': maximum allowed function calls; passed to the least squares fitter
        'fitting_tolerance': target difference between the fit and the data
        'symmetric': flag to signal if a symmetric gaussian is desired (i.e. sigma_x == sigma_y)
        'background': flag to signal if a background constant is desired
        'movement_allowance': percent of movement allowed (on all parameters except x and y positions

    """

    def __init__(self, atom_grp, fitting_parms):
        # we should do some initial checks here to ensure the data is at the correct stage for atom fitting
        print('Initializing Gauss Fit')
        # check that the data is appropriate (does nothing yet)
        self.check_data(atom_grp)

        # set dtypes
        self.atom_coeff_dtype = np.dtype([('type', np.float32),
                                          ('amplitude', np.float32),
                                          ('x', np.float32),
                                          ('y', np.float32),
                                          ('sigma_x', np.float32),
                                          ('sigma_y', np.float32),
                                          ('theta', np.float32),
                                          ('background', np.float32)])

        self.motif_coeff_dtype = np.dtype([('amplitude', np.float32),
                                           ('x', np.float32),
                                           ('y', np.float32),
                                           ('sigma_x', np.float32),
                                           ('sigma_y', np.float32),
                                           ('theta', np.float32),
                                           ('background', np.float32)])

        # initialize some variables
        self.atom_grp = atom_grp
        self.cropped_clean_image = self.atom_grp['Cropped_Clean_Image'][()]
        self.h5_guess = self.atom_grp['Guess_Positions']
        self.fitting_parms = fitting_parms
        self.win_size = self.atom_grp.attrs['motif_win_size']
        half_wind = int(self.win_size * 0.5)
        self.motif_centers = self.atom_grp['Motif_Centers'][:] - half_wind  # << correction for double cropped image
        self.psf_width = self.atom_grp.attrs['psf_width']

        # grab the initial guesses
        self.all_atom_guesses = np.transpose(np.vstack((self.h5_guess['x'],
                                                        self.h5_guess['y'],
                                                        self.h5_guess['type'])))

        self.num_atoms = self.all_atom_guesses.shape[0]  # number of atoms

        # build distance matrix
        pos_vec = self.all_atom_guesses[:, 0] + 1j * self.all_atom_guesses[:, 1]

        pos_mat1 = np.tile(np.transpose(np.atleast_2d(pos_vec)), [1, self.num_atoms])
        pos_mat2 = np.transpose(pos_mat1)
        d_mat = np.abs(pos_mat2 - pos_mat1)  # matrix of distances between all atoms
        # sort the distance matrix and keep only the atoms within the nearest neighbor limit
        neighbor_dist_order = np.argsort(d_mat)

        self.num_nearest_neighbors = self.fitting_parms['num_nearest_neighbors']

        # neighbor dist order has the (indices of the) neighbors for each atom sorted by distance
        self.closest_neighbors_mat = neighbor_dist_order[:, 1:self.num_nearest_neighbors + 1]

        # find which atoms are at the centers of the motifs
        self.center_atom_vec = (self.motif_centers[:, 0]) + (1j * self.motif_centers[:, 1])
        self.center_atom_dists = [np.abs(pos_vec - item) for item in self.center_atom_vec]
        self.center_atom_sorted = np.argsort(self.center_atom_dists)
        self.center_atom_indices = self.center_atom_sorted[:, 0]

    def fit_atom_positions_parallel(self, plot_results=True, num_cores=None):
        """
        Fits the positions of N atoms in parallel

        Parameters
        ----------
        plot_results : optional boolean (default is True)
            Specifies whether to output a visualization of the fitting results

        num_cores : unsigned int (Optional. Default = available logical cores - 2)
            Number of cores to compute with

        Creates guess_dataset and fit_dataset with the results.

        Returns
        -------

        fit_dataset: NxM numpy array of tuples where N is the number of atoms fit and M is the number of nearest
            neighbors considered. Each tuple contains the converged values for each gaussian.
            The value names are stored in the dtypes.
        """

        t_start = tm.time()
        if num_cores is None:
            num_cores = recommendCores(self.num_atoms, requested_cores=num_cores, lengthy_computation=False)

        print('Setting up guesses')
        self.guess_parms = []
        for i in range(self.num_atoms):
            self.guess_parms.append(self.do_guess(i))

        print('Fitting...')
        if num_cores > 1:
            pool = mp.Pool(processes=num_cores)
            parm_list = itt.izip(self.guess_parms, itt.repeat(self.fitting_parms))
            chunk = int(self.num_atoms / num_cores)
            jobs = pool.imap(do_fit, parm_list, chunksize=chunk)
            self.fitting_results = [j for j in jobs]
            pool.close()
        else:
            parm_list = itt.izip(self.guess_parms, itt.repeat(self.fitting_parms))
            self.fitting_results = [do_fit(parm) for parm in parm_list]

        print ('Finalizing datasets...')
        self.guess_dataset = np.zeros(shape=(self.num_atoms, self.num_nearest_neighbors + 1), dtype=self.atom_coeff_dtype)
        self.fit_dataset = np.zeros(shape=self.guess_dataset.shape, dtype=self.guess_dataset.dtype)

        for atom_ind, single_atom_results in enumerate(self.fitting_results):
            types = np.hstack((self.h5_guess['type'][atom_ind], [self.h5_guess['type'][neighbor] for neighbor in self.closest_neighbors_mat[atom_ind]]))
            atom_data = np.hstack((np.vstack(types), single_atom_results))
            atom_data = [tuple(element) for element in atom_data]
            self.fit_dataset[atom_ind] = atom_data

            single_atom_guess = self.guess_parms[atom_ind]
            atom_guess_data = np.hstack((np.vstack(types), single_atom_guess[1]))
            atom_guess_data = [tuple(element) for element in atom_guess_data]
            self.guess_dataset[atom_ind] = atom_guess_data

        tot_time = np.round(tm.time() - t_start)
        print('Took {} sec to find {} atoms with {} cores'.format(tot_time, len(self.fitting_results), num_cores))

        # if plotting is desired
        if plot_results:
            fig, axis = plt.subplots(figsize=(14, 14))
            axis.hold(True)
            axis.imshow(self.cropped_clean_image, interpolation='none', cmap="gray")
            axis.scatter(self.guess_dataset[:, 0]['y'], self.guess_dataset[:, 0]['x'], color='yellow', label='Guess')
            axis.scatter(self.fit_dataset[:, 0]['y'], self.fit_dataset[:, 0]['x'], color='red', label='Fit')
            axis.legend()
            fig.tight_layout()
            fig.show()

        return self.fit_dataset

    def do_guess(self, atom_ind, initial_motifs=False):
        """
        Fits the position of a single atom.

        Parameters
        ----------
        atom_ind : int
            The index of the atom to generate guess parameters for

        initial_motifs : optional boolean (default is False)
            Specifies whether we are generating guesses for the initial motifs. Subsequent guesses
            have the advantage of the fits from the motifs and will be much better starting values.

        Returns
        -------
        atom_ind : int
            The index of the atom to generate guess parameters for

        coef_guess_mat : 2D numpy array
            Initial guess parameters for all the gaussians.

        fit_region : 2D numpy array
            The fit region cropped from the image

        s1 and s2 : 2D numpy arrays
            The required input for the X and Y parameters of gauss2d

        lb_mat and ub_mat : 2D numpy arrays
            The lower and upper bounds for the fitting.
        """

        fit_region_size = self.fitting_parms['fit_region_size']
        movement_allowance = self.fitting_parms['movement_allowance']
        position_range = self.fitting_parms['position_range']

        # start writing down initial guesses
        x_center_atom = self.h5_guess['x'][atom_ind]
        y_center_atom = self.h5_guess['y'][atom_ind]
        x_neighbor_atoms = [self.h5_guess['x'][self.closest_neighbors_mat[atom_ind][i]] for i in
                            range(self.num_nearest_neighbors)]
        y_neighbor_atoms = [self.h5_guess['y'][self.closest_neighbors_mat[atom_ind][i]] for i in
                            range(self.num_nearest_neighbors)]

        # select the window we're going to be fitting
        x_range = slice(max(int(np.round(x_center_atom - fit_region_size)), 0),
                        min(int(np.round(x_center_atom + fit_region_size)),
                            self.cropped_clean_image.shape[0]))
        y_range = slice(max(int(np.round(y_center_atom - fit_region_size)), 0),
                        min(int(np.round(y_center_atom + fit_region_size)),
                            self.cropped_clean_image.shape[1]))
        fit_region = self.cropped_clean_image[x_range, y_range]

        # define x and y fitting range
        s1, s2 = np.meshgrid(range(x_range.start, x_range.stop),
                             range(y_range.start, y_range.stop))

        # guesses are different if we're fitting the initial windows
        if initial_motifs:
            # If true, we need to generate more crude guesses
            # for the initial motif window fitting.
            # Once these have been fit properly they will act
            # as the starting point for future guesses.

            # put the initial guesses into the proper form
            x_guess = np.hstack((x_center_atom, x_neighbor_atoms))
            y_guess = np.hstack((y_center_atom, y_neighbor_atoms))
            sigma_x_center_atom = self.fitting_parms['sigma_guess']
            sigma_y_center_atom = self.fitting_parms['sigma_guess']
            sigma_x_neighbor_atoms = [self.fitting_parms['sigma_guess'] for i in
                                      range(self.num_nearest_neighbors)]
            sigma_y_neighbor_atoms = [self.fitting_parms['sigma_guess'] for i in
                                      range(self.num_nearest_neighbors)]
            theta_center_atom = 0
            theta_neighbor_atoms = np.zeros(self.num_nearest_neighbors)
            background_center_atom = np.min(fit_region)

            # The existence of a background messes up a straight forward gaussian amplitude guess,
            # so we add/subtract the background value from the straight forward guess depending
            # on if the background is positive or negative.
            if np.min(fit_region) < 0:
                a_guess = self.cropped_clean_image[
                              np.rint(x_guess).astype(int), np.rint(y_guess).astype(int)] - background_center_atom
            else:
                a_guess = self.cropped_clean_image[
                              np.rint(x_guess).astype(int), np.rint(y_guess).astype(int)] + background_center_atom

            sigma_x_guess = np.hstack((sigma_x_center_atom, sigma_x_neighbor_atoms))
            sigma_y_guess = np.hstack((sigma_y_center_atom, sigma_y_neighbor_atoms))
            theta_guess = np.hstack((theta_center_atom, theta_neighbor_atoms))
            background_guess = np.hstack([background_center_atom for num in range(
                self.num_nearest_neighbors + 1)])  # we will only need one background
            coef_guess_mat = np.transpose(np.vstack((a_guess, x_guess, y_guess, sigma_x_guess, sigma_y_guess,
                                                     theta_guess, background_guess)))
        else:
            # otherwise better guesses are assumed to exist
            motif_type = self.h5_guess['type'][atom_ind]
            coef_guess_mat = np.copy(self.motif_converged_parms[motif_type])
            coef_guess_mat[:, 1] = self.h5_guess['x'][atom_ind] + coef_guess_mat[:, 1]
            coef_guess_mat[:, 2] = self.h5_guess['y'][atom_ind] + coef_guess_mat[:, 2]

        # Choose upper and lower bounds for the fitting
        #
        # Address negatives first
        lb_a = []
        ub_a = []
        for item in coef_guess_mat[:, 0]:  # amplitudes

            if item < 0:
                lb_a.append(item + item * movement_allowance)
                ub_a.append(item - item * movement_allowance)
            else:
                lb_a.append(item - item * movement_allowance)
                ub_a.append(item + item * movement_allowance)

        lb_background = []
        ub_background = []
        for item in coef_guess_mat[:, 6]:  # background
            if item < 0:
                lb_background.append(item + item * movement_allowance)
                ub_background.append(item - item * movement_allowance)
            else:
                lb_background.append(item - item * movement_allowance)
                ub_background.append(item + item * movement_allowance)

        # Set up upper and lower bounds:
        lb_mat = [lb_a,                                                              # amplitude
                  coef_guess_mat[:, 1] - position_range,                             # x position
                  coef_guess_mat[:, 2] - position_range,                             # y position
                  [np.max([0, value - value * movement_allowance]) for value in coef_guess_mat[:, 3]],  # sigma x
                  [np.max([0, value - value * movement_allowance]) for value in coef_guess_mat[:, 4]],  # sigma y
                  coef_guess_mat[:, 5] - 2 * 3.14159,                                # theta
                  lb_background]                                                     # background

        ub_mat = [ub_a,                                                              # amplitude
                  coef_guess_mat[:, 1] + position_range,                             # x position
                  coef_guess_mat[:, 2] + position_range,                             # y position
                  coef_guess_mat[:, 3] + coef_guess_mat[:, 3] * movement_allowance,  # sigma x
                  coef_guess_mat[:, 4] + coef_guess_mat[:, 4] * movement_allowance,  # sigma y
                  coef_guess_mat[:, 5] + 2 * 3.14159,                                # theta
                  ub_background]                                                     # background

        lb_mat = np.transpose(lb_mat)
        ub_mat = np.transpose(ub_mat)

        check_bounds = False
        if check_bounds:
            for i, item in enumerate(coef_guess_mat):
                for j, value in enumerate(item):
                    if lb_mat[i][j] > value or ub_mat[i][j] < value:
                        print('Atom number: {}'.format(atom_ind))
                        print('Guess: {}'.format(item))
                        print('Lower bound: {}'.format(lb_mat[i]))
                        print('Upper bound: {}'.format(ub_mat[i]))
                        print('dtypes: {}'.format(self.atom_coeff_dtype.names))
                        raise ValueError('{} guess is out of bounds'.format(self.atom_coeff_dtype.names[j]))

        return atom_ind, coef_guess_mat, fit_region, s1, s2, lb_mat, ub_mat

    def check_data(self, atom_grp):
        # some data checks here
        try:
            img = atom_grp['Cropped_Clean_Image']
        except KeyError:
            raise KeyError('The data \'Cropped_Clean_Image\' must exist before fitting')

        try:
            guesses = atom_grp['Guess_Positions']
        except KeyError:
            raise KeyError('The data \'Guess_Positions\' must exist before fitting')

        try:
            motifs = atom_grp['Motif_Centers']
        except KeyError:
            raise KeyError('The data \'Motif_Centers\' must exist before fitting')

        if np.shape(img)[0] < 1 or np.shape(img)[1] < 1:
            raise Exception('\'Cropped_Clean_Image\' data must have two dimensions with lengths greater than one')

        if len(guesses) < 1:
            raise Exception('\'Guess_Positions\' data length must be greater than one')

        if len(guesses[0]) < 3:
            raise Exception('\'Guess_Positions\' data must have at least three values for each entry: '
                            'type, x position, and y position')

        if motifs.shape[0] < 1:
            raise Exception('\'Motif_Centers\' data must contain at least one motif')

        if motifs.shape[1] != 2:
            raise Exception('\'Motif_Centers\' data is expected to have a shape of (n, 2). '
                            'The second dimension is not 2.')

    def write_to_disk(self):
        """
        Writes the gaussian fitting results to disk

        Parameters
        ----------

            None

        Returns
        -------

            Returns the atom parent group containing the original data and the newly written data:
                Gaussian_Guesses
                Gaussian_Fits
                Motif_Guesses
                Motif_Fits
                Nearest_Neighbor_Indices
        """

        ds_atom_guesses = MicroDataset('Gaussian_Guesses', data=self.guess_dataset)
        ds_atom_fits = MicroDataset('Gaussian_Fits', data=self.fit_dataset)
        ds_motif_guesses = MicroDataset('Motif_Guesses', data=self.motif_guess_dataset)
        ds_motif_fits = MicroDataset('Motif_Fits', data=self.motif_converged_dataset)
        ds_nearest_neighbors = MicroDataset('Nearest_Neighbor_Indices', data=self.closest_neighbors_mat, dtype=np.uint32)
        dgrp_atom_finding = MicroDataGroup(self.atom_grp.name.split('/')[-1], parent=self.atom_grp.parent.name)
        dgrp_atom_finding.attrs = self.fitting_parms
        dgrp_atom_finding.addChildren([ds_atom_guesses, ds_atom_fits, ds_motif_guesses, ds_motif_fits, ds_nearest_neighbors])

        hdf = ioHDF5(self.atom_grp.file)
        h5_atom_refs = hdf.writeData(dgrp_atom_finding)
        hdf.flush()
        return self.atom_grp

    def fit_motif(self, plot_results=True):
        '''
        Parameters
        ----------
            plot_results: boolean (default = True)
                Flag to specify whether a result summary should be plotted

        Returns
        -------
            motif_converged_dataset: NxM numpy array of tuples where N is the number of motifs and M is the number
                of nearest neighbors considered. Each tuple contains the converged parameters for a gaussian fit to
                an atom in a motif window.
        '''


        self.motif_guesses = []
        self.motif_parms = []
        self.motif_converged_parms = []
        self.fit_motifs = []
        fit_region = []

        # generate final dataset forms
        self.motif_guess_dataset = np.zeros(shape=(self.motif_centers.shape[0], self.num_nearest_neighbors + 1),
                                            dtype=self.motif_coeff_dtype)
        self.motif_converged_dataset = np.zeros(shape=(self.motif_centers.shape[0], self.num_nearest_neighbors + 1),
                                                dtype=self.motif_coeff_dtype)

        for motif in range(len(self.motif_centers)):
            # get guesses
            self.motif_parms.append(self.do_guess(self.center_atom_indices[motif], initial_motifs=True))

            # pull out parameters for generating the gaussians
            coef_guess_mat = self.motif_parms[motif][1]
            s1 = self.motif_parms[motif][3].T
            s2 = self.motif_parms[motif][4].T
            fit_region.append(self.motif_parms[motif][2])

            # put guesses into final dataset form
            self.motif_guess_dataset[motif] = [tuple(element) for element in coef_guess_mat]

            # store the guess results for plotting
            self.motif_guesses.append(gauss2d(s1, s2, *coef_guess_mat, **self.fitting_parms))

            # fit the motif with num_nearest_neighbors + 1 gaussians
            parm_list = [self.motif_parms[motif], self.fitting_parms]
            fitting_results = do_fit(parm_list)

            # store the converged results
            self.motif_converged_parms.append(fitting_results)
            self.motif_converged_dataset[motif] = [tuple(element) for element in fitting_results]

            # store the images of the converged gaussians
            self.fit_motifs.append(gauss2d(s1, s2, *fitting_results, **self.fitting_parms))

            # calculate the relative atom positions (instead of absolute)
            fitting_results[:, 1] = fitting_results[:, 1] - self.motif_centers[motif][0]
            fitting_results[:, 2] = fitting_results[:, 2] - self.motif_centers[motif][1]

        # plot results if desired
        if plot_results:

            # initialize the figure
            fig, axes = plt.subplots(ncols=3, nrows=len(self.motif_centers), figsize=(14, 6 * len(self.motif_centers)))

            for i, ax_row in enumerate(np.atleast_2d(axes)):
                # plot the original windows
                ax_row[0].imshow(fit_region[i], interpolation='none',
                                 cmap=plot_utils.cmap_jet_white_center())
                ax_row[0].set_title('Original Window')

                # plot the initial guess windows
                ax_row[1].imshow(self.motif_guesses[i], interpolation='none',
                                 cmap=plot_utils.cmap_jet_white_center())
                ax_row[1].set_title('Initial Gaussian Guesses')

                # plot the converged gaussians
                ax_row[2].imshow(self.fit_motifs[i], interpolation='none',
                                 cmap=plot_utils.cmap_jet_white_center())
                ax_row[2].set_title('Converged Gaussians')

            fig.show()

        return self.motif_converged_dataset


# if __name__=='__main__':
#     file_name = r"C:\Users\o2d\Documents\pycroscopy\\test_scripts\\testing_gauss_fit\image 04.h5"
#     folder_path, file_path = os.path.split(file_name)
#
#     file_base_name, file_extension = file_name.rsplit('.')
#     h5_file = h5py.File(file_name, mode='r+')
#     # look at the data tree in the h5
#     '''
#     # define a small function called 'print_tree' to look at the folder tree structure
#     def print_tree(parent):
#         print(parent.name)
#         if isinstance(parent, h5py.Group):
#             for child in parent:
#                 print_tree(parent[child])
#     '''
#
#     #print('Datasets and datagroups within the file:')
#     file_handle = h5_file
#     #print_tree(file_handle)
#
#     cropped_clean_image = h5_file['/Measurement_000/Channel_000/Raw_Data-Windowing_000/Image_Windows-SVD_000/U-Cluster_000/Labels-Atom_Finding_000/Cropped_Clean_Image']
#     atom_grp = cropped_clean_image.parent
#     guess_params = atom_grp['Guess_Positions']
#
#     num_nearest_neighbors = 4
#     psf_width = atom_grp.attrs['psf_width']
#     win_size = atom_grp.attrs['motif_win_size']
#
#     fitting_parms = {'fit_region_size': win_size * 0.5,  # region to consider when fitting
#                      'num_nearest_neighbors': num_nearest_neighbors,
#                      'sigma_guess': 3, # starting guess for gaussian standard deviation
#                      'position_range': win_size / 4,# range that the fitted position can go from initial guess position[pixels]
#                      'max_function_evals': 100,
#                      'fitting_tolerance': 1E-4,
#                      'symmetric': True,
#                      'background': True,
#                      'movement_allowance': 5.0} # percent of movement allowed (on some parameters)
#
#     foo = Gauss_Fit(atom_grp, fitting_parms)
#
#



"""
Created on Jun 16, 2016

@author: Chris Smith -- csmith55@utk.edu
"""

from __future__ import division, print_function, absolute_import
import os
from multiprocessing import cpu_count
from warnings import warn
import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import leastsq
from scipy.signal import blackman
from sklearn.utils import gen_batches

from ..io.hdf_utils import getH5DsetRefs, copyAttributes, linkRefs, findH5group, calc_chunks, link_as_main, \
    check_for_old
from ..io.io_hdf5 import ioHDF5
from ..io.io_utils import getAvailableMem
from ..io.microdata import MicroDataGroup, MicroDataset
from ..io.translators.utils import get_position_slicing, make_position_mat, get_spectral_slicing
from .svd_utils import _get_component_slice

# windata32 = np.dtype([('Image Data', np.float32)])
# absfft32 = np.dtype([('FFT Magnitude', np.float32)])
# winabsfft32 = np.dtype([('Image Data', np.float32), ('FFT Magnitude', np.float32)])
# wincompfft32 = np.dtype([('Image Data', np.float32), ('FFT Real', np.float32), ('FFT Imag', np.float32)])
windata32 = np.dtype({'names': ['Image Data'],
                      'formats': [np.float32]})
absfft32 = np.dtype({'names': ['FFT Magnitude'],
                     'formats': [np.float32]})
winabsfft32 = np.dtype({'names': ['Image Data', 'FFT Magnitude'],
                        'formats': [np.float32, np.float32]})
wincompfft32 = np.dtype({'names': ['Image Data', 'FFT Real', 'FFT Imag'],
                         'formats': [np.float32, np.float32, np.float32]})

class ImageWindow(object):
    """
    This class will handle the reading of a raw image file, creating windows from it, and writing those
    windows to an HDF5 file.
    """

    def __init__(self, h5_main, max_RAM_mb=1024, cores=None, reset=True, **image_args):
        """
        Setup the image windowing

        Parameters
        ----------
        h5_main : h5py.Dataset
            HDF5 Dataset containing the data to be windowed
        max_RAM_mb : int, optional
            integer maximum amount of ram, in Mb, to use in windowing
            Default 1024
        cores : int, optional
            integer number of [logical] CPU cores to use in windowing
            Defualt None, use number of available cores minus 2
        reset : Boolean, optional
            should all data in the hdf5 file be deleted

        """
        self.h5_file = h5_main.file
        self.hdf = ioHDF5(h5_main.file)
        
        # Ensuring that at least one core is available for use / 2 cores are available for other use
        max_cores = max(1, cpu_count()-2)
#         print 'max_cores',max_cores         
        if cores is not None: 
            cores = min(round(abs(cores)), max_cores)
        else:
            cores = max_cores            
        self.cores = int(cores)        

        self.max_memory = min(max_RAM_mb*1024**2, 0.75*getAvailableMem())
        if self.cores != 1:
            self.max_memory = int(self.max_memory/2)

        '''
        Initialize class variables to None
        '''
        self.h5_raw = h5_main
        self.h5_norm = None
        self.h5_wins = None
        self.h5_clean = None
        self.h5_noise = None
        self.h5_fft_clean = None
        self.h5_fft_noise = None

    def do_windowing(self, win_x=None, win_y=None, win_step_x=1, win_step_y=1,
                     win_fft=None, *args, **kwargs):
        """
        Extract the windows from the normalized image and write them to the file

        Parameters
        ----------
        win_x : int, optional
            size of the window, in pixels, in the horizontal direction
            Default None, a guess will be made based on the FFT of the image
        win_y : int, optional
            size of the window, in pixels, in the vertical direction
            Default None, a guess will be made based on the FFT of the image
        win_step_x : int, optional
            step size, in pixels, to take between windows in the horizontal direction
            Default 1
        win_step_y : int, optional
            step size, in pixels, to take between windows in the vertical direction
            Default 1
        win_fft : str, optional
            What kind of fft should be stored with the windows.  Options are
            None - Only the window
            'abs' - Only the magnitude of the fft
            'data+abs' - The window and magnitude of the fft
            'data+complex' - The window and the complex fft
            Default None

        Returns
        -------
        h5_wins : HDF5 Dataset
            Dataset containing the flattened windows

        """
        h5_main = self.h5_raw

        if win_fft == 'data' or win_fft is None:
            win_fft = 'data'
            win_type = windata32
            win_func = self.win_data_func
        elif win_fft == 'abs':
            win_type = absfft32
            win_func = self.abs_fft_func
        elif win_fft == 'data+abs':
            win_type = winabsfft32
            win_func = self.win_abs_fft_func
        elif win_fft == 'data+complex':
            win_type = wincompfft32
            win_func = self.win_comp_fft_func
        else:
            warn('Invalid FFT option supplied.  Windowing will default to data only.')
            win_fft = 'data'
            win_type = windata32
            win_func = self.win_data_func

        '''
        If a window size has not been specified, obtain a guess value from 
        window_size_extract
        '''
        win_test, psf_width = self.window_size_extract(*args, **kwargs)
        if win_x is None:
            win_x = win_test
        if win_y is None:
            win_y = win_test

        image, h5_wins, win_pos_mat, have_old = self._setup_window_h5(h5_main, psf_width, win_fft, win_step_x,
                                                                      win_step_y, win_type, win_x, win_y)

        if have_old:
            self.h5_wins = h5_wins
            return h5_wins

        n_wins = win_pos_mat.shape[0]
        win_pix = win_x * win_y

        '''
        Create slice object from the positions
        '''
        win_slices = [[slice(x, x+win_x), slice(y, y+win_y)] for x, y in win_pos_mat]

        '''
        Calculate the size of a given batch that will fit in the available memory
        '''
        mem_per_win = win_x*win_y*h5_wins.dtype.itemsize
        if self.cores is None:
            free_mem = self.max_memory-image.size*image.itemsize
        else:
            free_mem = self.max_memory*2-image.size*image.itemsize
        batch_size = int(free_mem/mem_per_win)
        batch_slices = gen_batches(n_wins, batch_size)

        for ibatch, batch in enumerate(batch_slices):
            batch_wins = np.zeros([batch.stop-batch.start, win_pix], dtype=win_type)
            '''
            Read each slice and write it to the dataset
            '''
            for islice, this_slice in enumerate(win_slices[batch]):
                iwin = ibatch*batch_size+islice
                selected = iwin % np.rint(n_wins/10) == 0

                if selected:
                    per_done = np.rint(100*iwin/n_wins)
                    print('Windowing Image...{}% --pixels {}-{}, step # {}'.format(per_done,
                                                                                   (this_slice[0].start,
                                                                                    this_slice[1].start),
                                                                                   (this_slice[0].stop,
                                                                                    this_slice[1].stop),
                                                                                   islice))

                batch_wins[islice] = win_func(image[this_slice]).flatten()

            h5_wins[batch] = batch_wins
            self.hdf.flush()
        
        self.h5_wins = h5_wins
        
        return h5_wins

    @staticmethod
    def _check_win_parameters(h5_main, win_step_x, win_step_y, win_x, win_y):
        """
        Check the window parameters
        
        Parameters
        ----------
        h5_main : h5py.Dataset
            Dataset containing the Raw Image
        win_step_x : uint
            Step size in the x-direction between windows.
        win_step_y : uint
            Step size in the y-direction between windows.
        win_x : uint
            Size of the window in the x-direction.
        win_y : uint
            Size of the window in the y-direction.

        Returns
        -------
        image : numpy.ndarray
            Array containing the original image reshaped by position.
        win_step_x : uint
            Step size in the x-direction between windows.
        win_step_y : uint
            Step size in the y-direction between windows.
        win_x : uint
            Size of the window in the x-direction.
        win_y : uint
            Size of the window in the y-direction.
        
        """
        '''
        Get the position indices of h5_main and reshape the flattened image back
        '''
        try:
            h5_pos = h5_main.file[h5_main.attrs['Position_Indices']][()]
            x_pix = len(np.unique(h5_pos[:, 0]))
            y_pix = len(np.unique(h5_pos[:, 1]))

        except KeyError:
            '''
            Position Indices dataset does not exist
            Assume square image
            '''
            x_pix = np.int(np.sqrt(h5_main.size))
            y_pix = x_pix

        except:
            raise
        image = h5_main[()].reshape(x_pix, y_pix)
        '''
            Step size must be less than 1/4th the image size
            '''
        win_step_x = min(x_pix / 4, win_step_x)
        win_step_y = min(y_pix / 4, win_step_y)
        '''
            Prevent windows from being less that twice the step size and more than half the image size
            '''
        win_x = max(2 * win_step_x, min(x_pix, win_x))
        win_y = max(2 * win_step_y, min(y_pix, win_y))
        print('Optimal window size determined to be {wx}x{wy} pixels.'.format(wx=win_x, wy=win_y))

        return image, win_step_x, win_step_y, win_x, win_y

    def _setup_window_h5(self, h5_main, psf_width, win_fft, win_step_x, win_step_y, win_type, win_x,
                         win_y):
        """
        Setup the hdf5 group for the windows
        
        Parameters
        ----------
        h5_main : h5py.Dataset
            Dataset containing the Raw Image
        psf_width : uint
            psf_width???  Someone who knows what this is should fill it in 
        win_fft : 
        win_step_x : uint
            Step size in the x-direction between windows.
        win_step_y : uint
            Step size in the y-direction between windows.
        win_type
        win_x : uint
            Size of the window in the x-direction.
        win_y : uint
            Size of the window in the y-direction.

        Returns
        -------
        

        """

        image, win_step_x, win_step_y, win_x, win_y = self._check_win_parameters(h5_main,
                                                                                 win_step_x, win_step_y,
                                                                                 win_x, win_y)

        '''
        Build the Spectroscopic and Position Datasets 
        '''
        ds_pix_inds, ds_pix_vals, ds_pos_inds, ds_pos_vals, win_pos_mat = self._get_window_pos_spec(
            image, win_step_x, win_step_y, win_x, win_y)
        im_x, im_y = image.shape
        n_wins = win_pos_mat.shape[0]
        win_pix = win_x * win_y

        '''
        Calculate the chunk size
        '''
        win_chunks = calc_chunks([n_wins, win_pix], win_type.itemsize, unit_chunks=[1, win_pix])

        parent = h5_main.parent

        check_parameters = {'fft_mode': win_fft,
                            'psf_width': psf_width,
                            'win_x': win_x,
                            'win_y': win_y,
                            'win_step_x': win_step_x,
                            'win_step_y': win_step_y,
                            'image_x': im_x,
                            'image_y': im_y}
        basename = h5_main.name.split('/')[-1]

        old_group = check_for_old(h5_main, '-Windowing', check_parameters)

        if old_group is not None:
            old = True
            h5_wins = old_group['Image_Windows']

        else:
            old = False
            '''
            Create the Windows Dataset and Datagroup
            '''
            ds_windows = MicroDataset('Image_Windows',
                                      data=[],
                                      maxshape=[n_wins, win_pix],
                                      dtype=win_type,
                                      chunking=win_chunks,
                                      compression='gzip')

            ds_group = MicroDataGroup(basename + '-Windowing_', parent.name[1:])
            ds_group.addChildren([ds_windows, ds_pos_inds, ds_pix_inds,
                                  ds_pos_vals, ds_pix_vals])
            ds_group.attrs['win_x'] = win_x
            ds_group.attrs['win_y'] = win_y
            ds_group.attrs['win_step_x'] = win_step_x
            ds_group.attrs['win_step_y'] = win_step_y
            ds_group.attrs['image_x'] = im_x
            ds_group.attrs['image_y'] = im_y
            ds_group.attrs['psf_width'] = psf_width
            ds_group.attrs['fft_mode'] = win_fft
            image_refs = self.hdf.writeData(ds_group)

            '''
            Get the hdf5 objects for the windows and ancillary datasets
            '''
            h5_wins = getH5DsetRefs(['Image_Windows'], image_refs)[0]

            '''
            Link references to windowed dataset
            '''
            aux_ds_names = ['Position_Indices', 'Position_Values', 'Spectroscopic_Indices', 'Spectroscopic_Values']
            linkRefs(h5_wins, getH5DsetRefs(aux_ds_names, image_refs))

            self.hdf.flush()

        return image, h5_wins, win_pos_mat, old

    @staticmethod
    def _get_window_pos_spec(image, win_step_x, win_step_y, win_x, win_y):
        """
        Create the position and spectroscopic datasets for the windows.
        
        Parameters
        ----------
        image : numpy.ndarray
            Raw Image
        win_step_x : uint
            Step size in the x-direction between windows.
        win_step_y : uint
            Step size in the y-direction between windows.
        win_x : uint
            Size of the window in the x-direction.
        win_y : uint
            Size of the window in the y-direction.

        Returns
        -------
        ds_pix_inds : MicroDataset
            Spectroscopic Indices of the windows
        ds_pix_vals : MicroDataset
            Spectroscopic Values of the windows
        ds_pos_inds : MicroDataset
            Position Indices of the windows
        ds_pos_vals : MicroDataset
            Position Values of the windows
        win_pos_mat : numpy.ndarray
            Array containing the positions of the window origins

        """
        im_x, im_y = image.shape
        x_steps = np.arange(0, im_x - win_x + 1, win_step_x, dtype=np.uint32)
        y_steps = np.arange(0, im_y - win_y + 1, win_step_y, dtype=np.uint32)
        nx = len(x_steps)
        ny = len(y_steps)
        win_pos_mat = np.array([np.repeat(x_steps, ny), np.tile(y_steps, nx)], dtype=np.uint32).T
        win_pix_mat = make_position_mat([win_x, win_y]).T

        '''
        Set up the HDF5 Group and Datasets for the windowed data
        '''
        ds_pos_inds = MicroDataset('Position_Indices', data=win_pos_mat, dtype=np.int32)
        ds_pix_inds = MicroDataset('Spectroscopic_Indices', data=win_pix_mat, dtype=np.int32)
        ds_pos_vals = MicroDataset('Position_Values', data=win_pos_mat, dtype=np.float32)
        ds_pix_vals = MicroDataset('Spectroscopic_Values', data=win_pix_mat, dtype=np.float32)
        pos_labels = get_position_slicing(['Window Origin X', 'Window Origin Y'])
        ds_pos_inds.attrs['labels'] = pos_labels
        ds_pos_inds.attrs['units'] = ['pixel', 'pixel']
        ds_pos_vals.attrs['labels'] = pos_labels
        ds_pos_vals.attrs['units'] = ['pixel', 'pixel']
        pix_labels = get_spectral_slicing(['U', 'V'])
        ds_pix_inds.attrs['labels'] = pix_labels
        ds_pix_inds.attrs['units'] = ['pixel', 'pixel']
        ds_pix_vals.attrs['labels'] = pix_labels
        ds_pix_vals.attrs['units'] = ['pixel', 'pixel']

        return ds_pix_inds, ds_pix_vals, ds_pos_inds, ds_pos_vals, win_pos_mat

    @staticmethod
    def win_data_func(image):
        """
        Returns the input image in the `windata32` format

        Parameters
        ----------
        image : numpy.ndarray
            Windowed image to take the FFT of

        Returns
        -------
        windows : numpy.ndarray
            Array the image in the windata32 format
        """
        windows = np.empty_like(image, dtype=windata32)
        windows['Image Data'] = image

        return windows

    @staticmethod
    def abs_fft_func(image):
        """
        Take the 2d FFT of each window in `windows` and return in the proper form.

        Parameters
        ----------
        image : numpy.ndarray
            Windowed image to take the FFT of

        Returns
        -------
        windows : numpy.ndarray
            Array of the Magnitude of the FFT of each window for the input
            `image`

        """
        windows = np.empty_like(image, dtype=absfft32)
        windows['FFT Magnitude'] = np.abs(np.fft.fftshift(np.fft.fft2(image)))

        return windows

    @staticmethod
    def win_abs_fft_func(image):
        """
        Take the 2d FFT of each window in `windows` and return in the proper form.

        Parameters
        ----------
        image : numpy.ndarray
            Windowed image to take the FFT of

        Returns
        -------
        windows : numpy.ndarray
            Array of windows and the Magnitude of the FFT of each window for the input
            `image`

        """
        windows = np.empty_like(image, dtype=winabsfft32)
        windows['Image Data'] = image
        windows['FFT Magnitude'] = np.abs(np.fft.fftshift(np.fft.fft2(image)))

        return windows

    @staticmethod
    def win_comp_fft_func(image):
        """
        Take the 2d FFT of each window in `windows` and return in the proper form.

        Parameters
        ----------
        image : numpy.ndarray
            Windowed image to take the FFT of

        Returns
        -------
        windows : numpy.ndarray
            Array of windows and the FFT of each window for the input `image`

        """
        windows = np.empty_like(image, dtype=wincompfft32)
        windows['Image Data'] = image
        win_fft = np.fft.fftshift(np.fft.fft2(image))
        windows['FFT Real'] = win_fft.real
        windows['FFT Imag'] = win_fft.imag

        return windows

    def build_clean_image(self, h5_win=None):
        """
        Reconstructs the cleaned image from the windowed dataset

        Parameters
        ----------
        h5_win : HDF5 dataset , optional
            The windowed image to be reconstructed.

        Returns
        -------
        h5_clean : HDF5 dataset
            The cleaned image

        """
        if h5_win is None:
            if self.clean_wins is None:
                warn('You must clean the image before rebuilding it.')
                return
            h5_win = self.clean_wins
        
        '''
        Get basic windowing information from attributes of 
        h5_win
        '''
        im_x = h5_win.parent.attrs['image_x']
        im_y = h5_win.parent.attrs['image_y']
        win_x = h5_win.parent.attrs['win_x']
        win_y = h5_win.parent.attrs['win_y']
        win_step_x = h5_win.parent.attrs['win_step_x']
        win_step_y = h5_win.parent.attrs['win_step_x']             
        
        '''
        Calculate the steps taken to create original windows
        '''
        x_steps = np.arange(0, im_x-win_x+1, win_step_x)
        y_steps = np.arange(0, im_y-win_y+1, win_step_y)
        
        '''
        Initialize arrays to hold summed windows and counts for each position
        '''
        counts = np.zeros([im_x, im_y], np.uint8)
        accum = np.zeros([im_x, im_y], np.float32)
        
        nx = len(x_steps)
        ny = len(y_steps)
        n_wins = nx*ny
        
        '''
        Create slice object from the positions
        '''
        win_slices = [[slice(x, x+win_x), slice(y, y+win_y)] for x, y in np.array([np.tile(x_steps, nx),
                                                                                   np.repeat(y_steps, ny)]).T]
        
        '''
        Loop over all windows.  Increment counts for window positions and 
        add current window to total.
        '''
        ones = np.ones([win_x, win_y], dtype=counts.dtype)
        for islice, this_slice in enumerate(win_slices):
            selected = islice%np.rint(n_wins/10) == 0
            if selected:
                per_done = np.rint(100*(islice)/(n_wins))
                print('Reconstructing Image...{}% -- step # {}'.format(per_done,islice))
            counts[this_slice]+= ones
    
            accum[this_slice]+= h5_win[islice].reshape(win_x,win_y)

        clean_image = accum/counts
        
        clean_image[np.isnan(clean_image)] = 0
        
        clean_grp = MicroDataGroup('Cleaned_Image', h5_win.parent.name[1:])

        ds_clean = MicroDataset('Cleaned_Image', clean_image)
        
        clean_grp.addChildren([ds_clean])
        
        image_refs = self.hdf.writeData(clean_grp)
        self.hdf.flush()
        
        h5_clean = getH5DsetRefs(['Cleaned_Image'], image_refs)[0]
    
        self.h5_clean = h5_clean
    
        return h5_clean

    def clean_and_build(self, h5_win=None, components=None):
        """
        Rebuild the Image from the SVD results on the windows
        Optionally, only use components less than n_comp.

        Parameters
        ----------
        h5_win : hdf5 Dataset, optional
            dataset containing the windowed image which SVD was performed on
        components: {int, iterable of int, slice} optional
            Defines which components to keep

            Input Types
            integer : Components less than the input will be kept
            length 2 iterable of integers : Integers define start and stop of component slice to retain
            other iterable of integers or slice : Selection of component indices to retain

        Returns
        -------
        clean_wins : HDF5 Dataset
            the cleaned windows

        """

        if h5_win is None:
            if self.h5_wins is None:
                warn('You must perform windowing on an image followed by SVD on the window before you can clean it.')
                return
            h5_win = self.h5_wins
        elif 'Image Data' not in h5_win.dtype.names:
            warn('The windows must have the real space image data in them to rebuild.')
            return

        print('Cleaning the image by removing unwanted components.')

        comp_slice = _get_component_slice(components)

        '''
        Read the 1st n_comp components from the SVD results
        on h5_win
        '''
        win_name = h5_win.name.split('/')[-1]

        try:
            win_svd = findH5group(h5_win, 'SVD')[-1]

            h5_S = win_svd['S']
            h5_U = win_svd['U']
            h5_V = win_svd['V']

        except KeyError:
            warnstring = 'SVD Results for {dset} were not found in {file}.'.format(dset=win_name, file=self.image_path)
            warn(warnstring)
            return
        except:
            raise

        '''
        Get basic windowing information from attributes of
        h5_win
        '''
        im_x = h5_win.parent.attrs['image_x']
        im_y = h5_win.parent.attrs['image_y']
        win_x = h5_win.parent.attrs['win_x']
        win_y = h5_win.parent.attrs['win_y']
        win_step_x = h5_win.parent.attrs['win_step_x']
        win_step_y = h5_win.parent.attrs['win_step_x']

        '''
        Calculate the steps taken to create original windows
        '''
        x_steps = np.arange(0, im_x - win_x, win_step_x)
        y_steps = np.arange(0, im_y - win_y, win_step_y)

        '''
        Initialize arrays to hold summed windows and counts for each position
        '''
        counts = np.zeros([im_x, im_y], np.uint32)
        clean_image = np.zeros([im_x, im_y], np.float32)

        nx = len(x_steps)
        ny = len(y_steps)
        n_wins = nx * ny

        '''
        Create slice object from the positions
        '''
        h5_win_pos = h5_win.file[h5_win.attrs['Position_Indices']]
        win_slices = [[slice(x, x+win_x), slice(y, y+win_y)] for x, y in h5_win_pos]

        '''
        Loop over all windows.  Increment counts for window positions and
        add current window to total.
        '''
        ones = np.ones([win_x, win_y], dtype=counts.dtype)
        ds_V = np.dot(np.diag(h5_S[comp_slice]), h5_V['Image_Data'][comp_slice, :])

        for islice, this_slice in enumerate(win_slices):
            if islice % np.rint(n_wins / 10) == 0:
                per_done = np.rint(100 * (islice) / (n_wins))
                print('Reconstructing Image...{}% -- step # {}'.format(per_done, islice))

            counts[this_slice] += ones

            this_win = np.dot(h5_U[islice, comp_slice], ds_V)

            clean_image[this_slice] += this_win.reshape(win_x, win_y)

        clean_image = np.divide(clean_image, counts)

        clean_image[np.isnan(clean_image)] = 0

        '''
        Calculate the removed noise and FFTs
        '''
        removed_noise = np.reshape(self.h5_raw, clean_image.shape)-clean_image

        fft_clean = np.fft.fft2(clean_image)
        fft_noise = np.fft.fft2(removed_noise)

        '''
        Create datasets for results, link them properly, and write them to file
        '''
        clean_grp = MicroDataGroup('Cleaned_Image_', win_svd.name[1:])
        ds_clean = MicroDataset('Cleaned_Image', clean_image.reshape(self.h5_raw.shape))
        ds_noise = MicroDataset('Removed_Noise', removed_noise.reshape(self.h5_raw.shape))
        ds_fft_clean = MicroDataset('FFT_Cleaned_Image', fft_clean.reshape(self.h5_raw.shape))
        ds_fft_noise = MicroDataset('FFT_Removed_Noise', fft_noise.reshape(self.h5_raw.shape))

        clean_grp.addChildren([ds_clean, ds_noise, ds_fft_clean, ds_fft_noise])

        if isinstance(comp_slice, slice):
            clean_grp.attrs['components_used'] = '{}-{}'.format(comp_slice.start, comp_slice.stop)
        else:
            clean_grp.attrs['components_used'] = comp_slice

        image_refs = self.hdf.writeData(clean_grp)
        self.hdf.flush()

        h5_clean = getH5DsetRefs(['Cleaned_Image'], image_refs)[0]
        h5_noise = getH5DsetRefs(['Removed_Noise'], image_refs)[0]
        h5_fft_clean = getH5DsetRefs(['FFT_Cleaned_Image'], image_refs)[0]
        h5_fft_noise = getH5DsetRefs(['FFT_Removed_Noise'], image_refs)[0]

        copyAttributes(self.h5_raw, h5_clean, skip_refs=False)
        copyAttributes(self.h5_raw, h5_noise, skip_refs=False)
        copyAttributes(self.h5_raw, h5_fft_clean, skip_refs=False)
        copyAttributes(self.h5_raw, h5_fft_noise, skip_refs=False)

        self.h5_clean = h5_clean
        self.h5_noise = h5_noise

        return h5_clean

    def clean_and_build_batch(self, h5_win=None, components=None):
        """
        Rebuild the Image from the SVD results on the windows
        Optionally, only use components less than n_comp.

        Parameters
        ----------
        h5_win : hdf5 Dataset, optional
            dataset containing the windowed image which SVD was performed on
        components : {int, iterable of int, slice} optional
            Defines which components to keep
            Default - None, all components kept

            Input Types
            integer : Components less than the input will be kept
            length 2 iterable of integers : Integers define start and stop of component slice to retain
            other iterable of integers or slice : Selection of component indices to retain

        Returns
        -------
        clean_wins : HDF5 Dataset
            the cleaned windows

        """

        if h5_win is None:
            if self.h5_wins is None:
                warn('You must perform windowing on an image followed by SVD on the window before you can clean it.')
                return
            h5_win = self.h5_wins
        elif 'Image Data' not in h5_win.dtype.names:
            warn('The windows must have the real space image data in them to rebuild.')
            return

        print('Cleaning the image by removing unwanted components.')

        comp_slice = _get_component_slice(components)

        '''
        Read the 1st n_comp components from the SVD results
        on h5_win
        '''
        win_name = h5_win.name.split('/')[-1]

        try:
            win_svd = findH5group(h5_win, 'SVD')[-1]

            h5_S = win_svd['S']
            h5_U = win_svd['U']
            h5_V = win_svd['V']

        except KeyError:
            warnstring = 'SVD Results for {dset} were not found in {file}.'.format(dset=win_name, file=self.image_path)
            warn(warnstring)
            return
        except:
            raise

        '''
        Get basic windowing information from attributes of
        h5_win
        '''
        im_x = h5_win.parent.attrs['image_x']
        im_y = h5_win.parent.attrs['image_y']
        win_x = h5_win.parent.attrs['win_x']
        win_y = h5_win.parent.attrs['win_y']

        '''
        Initialize arrays to hold summed windows and counts for each position
        '''
        counts = np.zeros([im_x, im_y], np.uint32)
        accum = np.zeros([im_x, im_y], np.float32)

        '''
        Create slice object from the positions
        '''
        ds_win_pos = h5_win.file[h5_win.attrs['Position_Indices']][()]
        win_slices = [[slice(x, x+win_x), slice(y, y+win_y)] for x, y in ds_win_pos]
        n_wins = ds_win_pos.shape[0]
        '''
        Create a matrix to add when counting.
        h5_V is usually small so go ahead and take S.V
        '''
        ones = np.ones([win_x, win_y], dtype=counts.dtype)
        ds_V = np.dot(np.diag(h5_S[comp_slice]), h5_V['Image Data'][comp_slice, :])

        '''
        Calculate the size of a given batch that will fit in the available memory
        '''
        mem_per_win = ds_V.itemsize*ds_V.shape[1]
        if self.cores is None:
            free_mem = self.max_memory-ds_V.size*ds_V.itemsize
        else:
            free_mem = self.max_memory*2-ds_V.size*ds_V.itemsize
        batch_size = int(free_mem/mem_per_win)
        batch_slices = gen_batches(n_wins, batch_size)

        print('Reconstructing in batches of {} windows.'.format(batch_size))

        '''
        Loop over all batches.  Increment counts for window positions and
        add current window to total.
        '''
        for ibatch, batch in enumerate(batch_slices):
            ds_U = h5_U[batch, comp_slice]
            batch_wins = np.dot(ds_U, ds_V).reshape([-1, win_x, win_y])
            del ds_U
            for islice, this_slice in enumerate(win_slices[batch]):
                iwin = ibatch*batch_size+islice
                if iwin % np.rint(n_wins / 10) == 0:
                    per_done = np.rint(100 * iwin / n_wins)
                    print('Reconstructing Image...{}% -- step # {}'.format(per_done, islice))

                counts[this_slice] += ones

                accum[this_slice] += batch_wins[islice]

        clean_image = np.divide(accum, counts)

        clean_image[np.isnan(clean_image)] = 0

        if h5_win.file.attrs['normalized']:
            '''
            Renormalize the cleaned image
            '''
            clean_image -= np.min(clean_image)
            clean_image = clean_image/np.max(clean_image)
        

        '''
        Calculate the removed noise and FFTs
        '''
        removed_noise = np.reshape(self.h5_raw, clean_image.shape)-clean_image
        blackman_window_rows = blackman(clean_image.shape[0])
        blackman_window_cols = blackman(clean_image.shape[1])
        fft_clean = np.fft.fft2(blackman_window_rows[:, np.newaxis]*clean_image*blackman_window_cols[np.newaxis, :])
        fft_noise = np.fft.fft2(blackman_window_rows[:, np.newaxis]*removed_noise*blackman_window_cols[np.newaxis, :])

        '''
        Create datasets for results, link them properly, and write them to file
        '''
        clean_grp = MicroDataGroup('Cleaned_Image_', win_svd.name[1:])
        ds_clean = MicroDataset('Cleaned_Image', clean_image.reshape(self.h5_raw.shape))
        ds_noise = MicroDataset('Removed_Noise', removed_noise.reshape(self.h5_raw.shape))
        ds_fft_clean = MicroDataset('FFT_Cleaned_Image', fft_clean.reshape(self.h5_raw.shape))
        ds_fft_noise = MicroDataset('FFT_Removed_Noise', fft_noise.reshape(self.h5_raw.shape))

        clean_grp.addChildren([ds_clean, ds_noise, ds_fft_clean, ds_fft_noise])

        if isinstance(comp_slice, slice):
            clean_grp.attrs['components_used'] = '{}-{}'.format(comp_slice.start, comp_slice.stop)
        else:
            clean_grp.attrs['components_used'] = comp_slice

        image_refs = self.hdf.writeData(clean_grp)
        self.hdf.flush()

        h5_clean = getH5DsetRefs(['Cleaned_Image'], image_refs)[0]
        h5_noise = getH5DsetRefs(['Removed_Noise'], image_refs)[0]
        h5_fft_clean = getH5DsetRefs(['FFT_Cleaned_Image'], image_refs)[0]
        h5_fft_noise = getH5DsetRefs(['FFT_Removed_Noise'], image_refs)[0]

        copyAttributes(self.h5_raw, h5_clean, skip_refs=False)
        copyAttributes(self.h5_raw, h5_noise, skip_refs=False)
        copyAttributes(self.h5_raw, h5_fft_clean, skip_refs=False)
        copyAttributes(self.h5_raw, h5_fft_noise, skip_refs=False)

        self.h5_clean = h5_clean
        self.h5_noise = h5_noise

        return h5_clean

    def clean_and_build_separate_components(self, h5_win=None, components=None):
        """
        Rebuild the Image from the SVD results on the windows
        Optionally, only use components less than n_comp.

        Parameters
        ----------
        h5_win : hdf5 Dataset, optional
            dataset containing the windowed image which SVD was performed on
        components : {int, iterable of int, slice} optional
            Defines which components to keep
            Default - None, all components kept
            \n
            Input Types
            integer : Components less than the input will be kept
            length 2 iterable of integers : Integers define start and stop of component slice to retain
            other iterable of integers or slice : Selection of component indices to retain

        Returns
        -------
        clean_wins : HDF5 Dataset
            the cleaned windows

        """

        if h5_win is None:
            if self.h5_wins is None:
                warn('You must perform windowing on an image followed by SVD on the window before you can clean it.')
                return
            h5_win = self.h5_wins
        elif 'Image Data' not in h5_win.dtype.names:
            warn('The windows must have the real space image data in them to rebuild.')
            return

        print('Cleaning the image by removing unwanted components.')
        comp_slice = _get_component_slice(components)

        '''
        Read the 1st n_comp components from the SVD results
        on h5_win
        '''
        win_name = h5_win.name.split('/')[-1]

        try:
            win_svd = findH5group(h5_win, 'SVD')[-1]

            h5_S = win_svd['S']
            h5_U = win_svd['U']
            h5_V = win_svd['V']

        except KeyError:
            warnstring = 'SVD Results for {dset} were not found in {file}.'.format(dset=win_name, file=self.image_path)
            warn(warnstring)
            return
        except:
            raise

        '''
        Get basic windowing information from attributes of
        h5_win
        '''
        im_x = h5_win.parent.attrs['image_x']
        im_y = h5_win.parent.attrs['image_y']
        win_x = h5_win.parent.attrs['win_x']
        win_y = h5_win.parent.attrs['win_y']

        '''
        Create slice object from the positions
        '''
        ds_win_pos = h5_win.file[h5_win.attrs['Position_Indices']][()]
        win_slices = [[slice(x, x+win_x), slice(y, y+win_y), slice(None)] for x, y in ds_win_pos]
        n_wins = len(ds_win_pos)

        '''
        Go ahead and take the dot product of S and V.  Get the number of components
        from the length of S
        '''
        ds_V = np.dot(np.diag(h5_S[comp_slice]), h5_V['Image Data'][comp_slice, :]).T
        num_comps = ds_V.shape[1]

        '''
        Initialize arrays to hold summed windows and counts for each position
        '''
        ones = np.ones([win_x, win_y, num_comps], dtype=np.uint32)
        counts = np.zeros([im_x, im_y, num_comps], dtype=np.uint32)
        clean_image = np.zeros([im_x, im_y, num_comps], dtype=np.float32)

        '''
        Calculate the size of a given batch that will fit in the available memory
        '''
        mem_per_win = ds_V.itemsize*(num_comps+ds_V.size)
        if self.cores is None:
            free_mem = self.max_memory-ds_V.size*ds_V.itemsize
        else:
            free_mem = self.max_memory/2-ds_V.size*ds_V.itemsize
        batch_size = int(free_mem/mem_per_win)
        if batch_size < 1:
            raise MemoryError('Not enough memory to perform Image Cleaning.')
        batch_slices = gen_batches(n_wins, batch_size)

        print('Reconstructing in batches of {} windows.'.format(batch_size))
        '''
        Loop over all batches.  Increment counts for window positions and
        add current window to total.
        '''
        for ibatch, batch in enumerate(batch_slices):
            ds_U = h5_U[batch, comp_slice]
            batch_wins = ds_U[:, None, :]*ds_V[None, :, :]
            for islice, this_slice in enumerate(win_slices[batch]):
                iwin = ibatch * batch_size + islice
                if iwin % np.rint(n_wins / 10) == 0:
                    per_done = np.rint(100 * iwin / n_wins)
                    print('Reconstructing Image...{}% -- step # {}'.format(per_done, iwin))

                counts[this_slice] += ones

                clean_image[this_slice] += batch_wins[islice].reshape(win_x, win_y, num_comps)

        del ds_U, ds_V

        clean_image /= counts
        del counts
        clean_image[np.isnan(clean_image)] = 0

        '''
        Create datasets for results, link them properly, and write them to file
        '''
        clean_grp = MicroDataGroup('Cleaned_Image_', win_svd.name[1:])

        clean_chunking = calc_chunks([im_x*im_y, num_comps],
                                     clean_image.dtype.itemsize)
        ds_clean = MicroDataset('Cleaned_Image',
                                data=clean_image.reshape(im_x*im_y, num_comps),
                                chunking=clean_chunking,
                                compression='gzip')

        clean_grp.addChildren([ds_clean])

        if isinstance(comp_slice, slice):
            clean_grp.attrs['components_used'] = '{}-{}'.format(comp_slice.start,
                                                                comp_slice.stop)
        else:
            clean_grp.attrs['components_used'] = comp_slice

        image_refs = self.hdf.writeData(clean_grp)
        self.hdf.flush()

        h5_clean = getH5DsetRefs(['Cleaned_Image'], image_refs)[0]
        h5_comp_inds = h5_clean.file[h5_V.attrs['Position_Indices']]
        h5_pos_inds = self.h5_file[self.h5_raw.attrs['Position_Indices']]
        h5_pos_vals = self.h5_file[self.h5_raw.attrs['Position_Values']]

        link_as_main(h5_clean, h5_pos_inds, h5_pos_vals, h5_comp_inds, h5_S)

        self.h5_clean = h5_clean

        return h5_clean

    def plot_clean_image(self, h5_clean=None, image_path=None, image_type='png',
                         save_plots=True, show_plots=False, cmap='gray'):
        """
        Plot the cleaned image stored in the HDF5 dataset h5_clean

        Parameters
        ----------
        h5_clean : HDF5 dataset, optional
            cleaned image to be plotted
        image_path : str, optional
            path to save cleaned image file
            Default None, '_clean' will be appened to the name of the input image
        image_type : str, optional
            image format to save the cleaned image as
            Default 'png', all formats recognized by matplotlib.pyplot.imsave
            are allowed
        save_plots : Boolean, pptional
            If true, the image will be saved to image_path
            with the extention specified by image_type
            Default True
        show_plots : Boolean, optional
            If true, the image will be displayed on the screen
            Default False
        cmap : str, optional
            matplotlib colormap string designation

        Returns
        -------
        clean_image : Axis_Image
            object holding the plot of the cleaned image

        """
        if h5_clean is None:
            if self.h5_clean is None:
                warn('You must clean an image before it can be plotted.')
                return
            h5_clean = self.h5_clean

        '''
        Get the position indices of h5_clean and reshape the flattened image back
        '''
        try:
            h5_pos = h5_clean.file[h5_clean.attrs['Position_Indices']][()]
            x_pix = len(np.unique(h5_pos[:, 0]))
            y_pix = len(np.unique(h5_pos[:, 1]))

        except KeyError:
            '''
        Position Indices dataset does not exist
        Assume square image
            '''
            x_pix = np.int(np.sqrt(h5_clean.size))
            y_pix = x_pix

        except:
            raise

        image = h5_clean[()].reshape(x_pix, y_pix)

        if save_plots:
            if image_path is None:
                image_dir, basename = os.path.split(self.h5_file.filename)
                basename, _ = os.path.splitext(basename)
                basename = basename+'_clean.'+image_type
                image_path = os.path.join(image_dir, basename)
            
            plt.imsave(image_path, image, format=image_type, cmap=cmap)

        clean_image = plt.imshow(image, cmap=cmap)
        if show_plots:
            plt.show()
        
        return clean_image


    def window_size_extract(self, num_peaks=2, save_plots=True, show_plots=False):
        """
        Take the normalized image and extract from it an optimal window size

        Parameters
        ----------
        num_peaks : int, optional
            number of peaks to use during least squares fit
            Default 2
        save_plots : Boolean, optional
            If True then a plot showing the quality of the fit will be
            generated and saved to disk.  Ignored if do_fit is false.
            Default True
        show_plots : Boolean, optional
            If True then a plot showing the quality of the fit will be
            generated and shown on screen.  Ignored if do_fit is false.
            Default False

        Returns
        -------
        window_size : int
            Optimal window size in pixels
        psf_width : int
            Estimate atom spacing in pixels

        """

        def __gauss_fit(p, x):
            """
            simple gaussian fitting function
            """
            a = p[0]
            s = p[1]

            g = a*np.exp(-(x/s)**2)

            return g

        def __gauss_chi(p, x, y):
            """
            Simple chi-squared fit
            """
            gauss = __gauss_fit(p, x)

            chi2 = ((y-gauss)/y)**2

            return chi2

        h5_main = self.h5_raw

        print('Determining appropriate window size from image.')
        '''
        Normalize the image
        '''
        immin = np.min(h5_main)
        immax = np.max(h5_main)
        image = np.float32(h5_main - immin) / (immax - immin)

        '''
        Reshape the image based on the position indices
        '''
        try:
            h5_pos = h5_main.file[h5_main.attrs['Position_Indices']][()]
            x_pix = len(np.unique(h5_pos[:, 0]))
            y_pix = len(np.unique(h5_pos[:, 1]))

        except KeyError:
            '''
            Position Indices dataset does not exist
            Assume square image
            '''
            x_pix = np.int(np.sqrt(h5_main.size))
            y_pix = x_pix

        except:
            raise
        image = image.reshape([x_pix, y_pix])

        '''
        Perform an fft on the normalize image 
        '''
        im_shape = np.min(image.shape)
        
        def __hamming(data):
            """
            Simple hamming filter
            """
            u, v = np.shape(data)
            u_vec = np.linspace(0, 1, u)
            v_vec = np.linspace(0, 1, v)
            u_mat, v_mat = np.meshgrid(u_vec, v_vec, indexing='ij')
            h_filter = np.multiply((1-np.cos(2*np.pi*u_mat)), (1-np.cos(2*np.pi*v_mat)))/4.0
            
            return np.multiply(data, h_filter)
        
        im2 = image-np.mean(image)
        fim = np.fft.fftshift(np.fft.fft2(__hamming(im2)))
        
        imrange = np.arange(-im_shape/2, im_shape/2)
        uu, vv = np.meshgrid(imrange, imrange)
        
        '''
        Find max at each radial distance from the center
        '''
        r_n = int(im_shape/4)
        r_min = 0
        r_max = im_shape/2
        r_vec = np.linspace(r_min, r_max, r_n, dtype=np.float32).transpose()
        
        r_mat = np.abs(uu+1j*vv)
        
        fimabs = np.abs(fim)
        fimabs_max = np.zeros(r_n-1)
        
        for k in range(r_n-1):
            r1 = r_vec[k]
            r2 = r_vec[k+1]
            r_ind = np.where((r_mat >= r1) & (r_mat <= r2) == True)
            fimabs_max[k] = np.max(fimabs[r_ind])

        r_vec = r_vec[:-1] + (r_max-r_min)/(r_n-1.0)/2.0
        
        '''
        Find local maxima
        '''
        count = 0
        local_max = []
        for k in range(1, fimabs_max.size-1):
            if fimabs_max[k-1] < fimabs_max[k] and fimabs_max[k] > fimabs_max[k+1]:
                count += 1
                local_max.append(k)
        
        '''
        Get points corresponding to local maxima
        '''
        r_loc_max_vec = r_vec[local_max]
        fimabs_loc_max_vec = fimabs_max[local_max]
        
        '''
        Remove points below the radius of the tallest peak
        '''
        fimabs_loc_max_ind = np.argmax(fimabs_loc_max_vec)
        fimabs_loc_max_vec = fimabs_loc_max_vec[fimabs_loc_max_ind:]
        r_loc_max_vec = r_loc_max_vec[fimabs_loc_max_ind:]

        '''
        Sort the peaks from largest to smallest
        ''' 
        sort_ind = np.argsort(fimabs_loc_max_vec)[::-1]
        fimabs_sort = fimabs_loc_max_vec[sort_ind]
        r_sort = r_loc_max_vec[sort_ind]

        '''
        Check to ensure there are at least 2 peaks left.
        '''
        if fimabs_sort.size == 1:
            warn('Only one peak in sorted vector.  Simple estimate being used.')
            window_size = im_shape / (r_sort[0] + 0.5)

            window_size = np.clip(window_size, 1, int(im_shape / 6.0))

            window_size = np.int(np.round(window_size * 2))

            gauss_guess = (2*fimabs_sort[0], r_sort[0])

            psf_width = im_shape/gauss_guess[1]/np.pi

            return window_size, psf_width

        '''
        Only use specified number of peaks
        '''
        fimabs_sort = fimabs_sort[:num_peaks]
        r_sort = r_sort[:num_peaks]
        
        '''
        Fit to a gaussian
        '''
        gauss_guess = (2*np.max(fimabs_sort), r_sort[0])

        fit_vec, pcov, info, errmsg, success = leastsq(__gauss_chi,
                                                       gauss_guess,
                                                       args=(r_sort, fimabs_sort),
                                                       full_output=1,
                                                       maxfev=250)

        psf_width = im_shape/fit_vec[1]/np.pi

        if save_plots or show_plots:
            guess_vec = __gauss_fit(gauss_guess, r_vec)
            fit_vec = __gauss_fit(fit_vec, r_vec)
            self.__plot_window_fit(r_vec, r_sort, fimabs_max, fimabs_sort,
                                   guess_vec, fit_vec, save_plots, show_plots)

        window_size = im_shape/(r_sort[0]+0.5)

        window_size = np.clip(window_size, 1, int(im_shape / 6.0))

        window_size = np.int(np.round(window_size*2))

        return window_size, psf_width

    def __plot_window_fit(self, r_vec, r_sort, fft_absimage, fft_abssort, guess, fit, save_plots=True, show_plots=False):
        """
        Generate a plot showing the quality of the least-squares fit to the peaks of the FFT of the image

        Parameters
        ----------
        r_vec : numpy array
            1D array of unsorted radii in pixels
        r_sort : numpy array
            1D array of the sorted radii
        fft_absimage : numpy array
            1D array of the absolute value of the FFT of the normalized image
        fft_abssort : numpy array
            1D array of FFT_absimage after being sorted to match r_sort
        guess :  numpy array
            1D array of the gaussian guess
        fit : numpy array
            1D array of the fitted gaussian
        save_plots : Boolean, optional
            If True then a plot showing the quality of the fit will be
            generated and saved to disk.
            Default True
        show_plots : Boolean, optional
            If True then a plot showing the quality of the fit will be
            generated and shown on screen.
            Default False

        Returns
        -------
        None

        """
        
        fig = plt.figure(figsize=[8, 8], tight_layout=True)
        plt1, = plt.semilogy(r_vec, fft_absimage, label='magnitude')
        plt2, = plt.semilogy(r_sort, fft_abssort, 'ro', label='chosen peaks')
        plt3, = plt.semilogy(r_vec, guess, 'g', label='guess')
        plt4, = plt.semilogy(r_vec, fit, 'r', label='fit')
        ax = fig.gca()
        ax.autoscale(tight=True)
        ax.set_xlim(left=0)
        ax.set_ylim(bottom=1)
        ax.set_xlabel('radius [pixels]')
        ax.set_ylabel('max magnitude')
        plt.legend(handles=[plt1, plt2, plt3, plt4])

        if save_plots:
            folder,filename = os.path.split(self.hdf.path)
            basename, junk = os.path.splitext(filename)
            
            plotname = '_'.join([basename, 'window_fit'])
            plotpath = os.path.join(folder, plotname+'.png')
            
            fig.savefig(plotpath, format='png')
        
        if show_plots:
            plt.show(fig)
            
        plt.close(fig)

def radially_average_correlation(data_mat, num_r_bin):
    """
    Calculates the radially average correlation functions for a given 2D image

    Parameters
    ----------
    data_mat : 2D real numpy array
        Image to analyze
    num_r_bin : unsigned int
        Number of spatial bins to analyze

    Returns
    --------
    a_mat : 2D real numpy array
        Noise spectrum of the image
    a_rad_avg_vec : 1D real numpy array
        Average value of the correlation as a function of feature size
    a_rad_max_vec : 1D real numpy array
        Maximum value of the correlation as a function of feature size
    a_rad_min_vec : 1D real numpy array
        Minimum value of the correlation as a function of feature size
    a_rad_std_vec : 1D real numpy array
        Standard deviation of the correlation as a function of feature size

    """
    x_size = data_mat.shape[0]
    y_size = data_mat.shape[1]

    x_mesh, y_mesh = np.meshgrid(np.linspace(-1, 1, x_size),
                                 np.linspace(-1, 1, y_size))
    r_vec = np.sqrt(x_mesh ** 2 + y_mesh ** 2).flatten()

    s_mat = (np.abs(np.fft.fftshift(np.fft.fft2(data_mat)))) ** 2
    a_mat = np.abs(np.fft.fftshift((np.fft.ifft2(s_mat))))

    min_a = np.min(a_mat)
    a_mat = a_mat - min_a
    max_a = np.max(a_mat)
    a_mat = a_mat / max_a

    a_vec = a_mat.flatten()

    # bin results based on r
    a_rad_avg_vec = np.zeros(num_r_bin)
    a_rad_max_vec = np.zeros(a_rad_avg_vec.shape)
    a_rad_min_vec = np.zeros(a_rad_avg_vec.shape)
    a_rad_std_vec = np.zeros(a_rad_avg_vec.shape)
    r_bin_vec = np.zeros(a_rad_avg_vec.shape)

    step = 1 / (num_r_bin * 1.0 - 1)
    for k, r_bin in enumerate(np.linspace(0, 1, num_r_bin)):
        b = np.where((r_vec < r_bin + step) * (r_vec > r_bin) == True)[0]

        if b.size == 0:
            a_rad_avg_vec[k] = np.nan
            a_rad_min_vec[k] = np.nan
            a_rad_max_vec[k] = np.nan
            a_rad_std_vec[k] = np.nan
        else:
            a_bin = a_vec[b]
            a_rad_avg_vec[k] = np.mean(a_bin)
            a_rad_min_vec[k] = np.min(a_bin)
            a_rad_max_vec[k] = np.max(a_bin)
            a_rad_std_vec[k] = np.std(a_bin)
        r_bin_vec[k] = r_bin + 0.5 * step

    return a_mat, a_rad_avg_vec, a_rad_max_vec, a_rad_min_vec, a_rad_std_vec



"""
Created on Nov 8, 2016

@author: Chris Smith -- csmith55@utk.edu
"""

from __future__ import division, print_function, absolute_import, unicode_literals

import array
import os

import numpy as np
from skimage.io import imread

from . import dm4reader
from .dm3_image_utils import parse_dm_header, imagedatadict_to_ndarray


def read_image(image_path, *args, **kwargs):
    """
    Read the image file at `image_path` into a numpy array

    Parameters
    ----------
    image_path : str
        Path to the image file

    Returns
    -------
    image : numpy.ndarray
        Array containing the image from the file `image_path`
    image_parms : dict
        Dictionary containing image parameters.  If image type does not have
        parameters then an empty dictionary is returned.

    """
    ext = os.path.splitext(image_path)[1]
    if ext == '.dm3':
        kwargs.pop('as_grey',None)
        return read_dm3(image_path, *args, **kwargs)
    elif ext == '.dm4':
        kwargs.pop('as_grey', None)
        return read_dm4(image_path, *args, **kwargs)
    elif ext == '.txt':
        return read_txt(image_path, *args, **kwargs), dict()
    else:
        # Set the as_grey argument to True is not already provided.
        kwargs['as_grey'] = (kwargs.pop('as_grey', True))
        return imread(image_path, *args, **kwargs), dict()


def read_dm3(image_path, get_parms=True):
    """
    Read an image from a dm3 file into a numpy array

    image_path : str
        Path to the image file
    get_parms : Boolean, optional
        Should the parameters from the dm3 file be returned
        Default True

    Returns
    -------
    image : numpy.ndarray
        Array containing the image from the file `image_path`

    """
    image_file = open(image_path, 'rb')
    dmtag = parse_dm_header(image_file)
    img_index = -1
    image = imagedatadict_to_ndarray(dmtag['ImageList'][img_index]['ImageData'])
    image_parms = dmtag['ImageList'][img_index]['ImageTags']

    if get_parms:
        image_parms = unnest_parm_dicts(image_parms)
    else:
        image_parms = dict()

    return image, image_parms

def unnest_parm_dicts(image_parms, prefix=''):
    """
    Parses the nested image parameter dictionary and converts it to a single
    level dictionary, prepending the name of inner dictionaries to their
    keys to denote level.

    Parameters
    ----------
    image_parms : dict

    Returns
    -------

    """
    new_parms = dict()
    for name in image_parms.keys():
        val = image_parms[name]
        # print 'name',name,'val',val
        name = '-'.join([prefix]+name.split()).strip('-')
        if isinstance(val, dict):
            new_parms.update(unnest_parm_dicts(val, name))
        elif isinstance(val, list) and isinstance(val[0], dict):
            for thing in val:
                new_parms.update(unnest_parm_dicts(thing, name))
        else:
            new_parms[name] = try_tag_to_string(val)

    return new_parms

def read_dm4(file_path, *args, **kwargs):
    """
    Read dm4 file

    Parameters
    ----------
    file_path : str
        Path to the file to be read

    Returns
    -------
    image_array : numpy.ndarray
        Image data from the file located at `file_path`
    file_parms : dict
        Dictionary of parameters read from the dm4 file

    """
    get_parms = kwargs.pop('get_parms', True)
    header = kwargs.pop('header', None)

    file_parms = dict()
    dm4_file = dm4reader.DM4File.open(file_path)
    if header is None:
        tags = dm4_file.read_directory()
        header = tags.named_subdirs['ImageList'].dm4_tag
        image_list = tags.named_subdirs['ImageList'].unnamed_subdirs
    else:
        dm4_file.hfile.seek(header.offset)
        image_list = dm4_file.read_directory(header)

    for image_dir in image_list:
        image_data_tag = image_dir.named_subdirs['ImageData']
        image_tag = image_data_tag.named_tags['Data']

        x_dim = dm4_file.read_tag_data(image_data_tag.named_subdirs['Dimensions'].unnamed_tags[0])
        y_dim = dm4_file.read_tag_data(image_data_tag.named_subdirs['Dimensions'].unnamed_tags[1])

        image_array = np.array(dm4_file.read_tag_data(image_tag), dtype=np.float32)
        image_array = np.reshape(image_array, (y_dim, x_dim))

    if get_parms:
        file_parms = parse_dm4_parms(dm4_file, tags, '')
        file_parms['Image_Tag'] = header

    return image_array, file_parms

def parse_dm4_parms(dm4_file, tag_dir, base_name=''):
    """
    Recursive function to trace the dictionary tree of the Image Data
    and build a single dictionary of all parameters

    Parameters
    ----------
    dm4_file : DM4File
        File object of the dm4 file to be parsed.

    tag_dir : dict
        Dictionary to be traced.  Has the following attributes:
            tag_dir.name : str
                Name of the directory
            tag_dir.dm4_tag : str
                Contents of the directory

    base_name : str
        Base name of parameters.  Tag and subdirectory names will be appended
        for named tags and subdirectories.  Unnamed ones will recieve a number.
        Default ''.  'Root' is automatically prepended to the name.

    Returns
    -------
    parm_dict : dict()
        Dictionary containing the name:value pairs of all parameters `dm4_file`

    """
    parm_dict = dict()

    '''
    Loop over named tags
    '''
    for name in tag_dir.named_tags.keys():
        '''
        Skip Data tags.  These will be handled elseware.
        '''
        if name == 'Data':
            continue
        tag_name = '_'.join([base_name, name.replace(' ', '_')])
        if base_name == '':
            tag_name = 'Root'+tag_name
        tag_data = dm4_file.read_tag_data(tag_dir.named_tags[name])

        '''
        See if we can convert the array into a string
        '''
        tag_data = try_tag_to_string(tag_data)
        parm_dict[tag_name] = tag_data

    '''
    Loop over unnamed tags
    '''
    for itag, tag in enumerate(tag_dir.unnamed_tags):
        tag_name = '_'.join([base_name, 'Tag_{:03d}'.format(itag)])
        if base_name == '':
            tag_name = 'Root'+tag_name

        tag_data = dm4_file.read_tag_data(tag)

        '''
        See if we can convert the array into a string
        '''
        tag_data = try_tag_to_string(tag_data)
        parm_dict[tag_name] = tag_data

    '''
    Loop over named subdirectories
    '''
    for name in tag_dir.named_subdirs.keys():
        dir_name = '_'.join([base_name, name.replace(' ', '_')])
        sub_dir = tag_dir.named_subdirs[name]
        if base_name == '':
            dir_name = 'Root'+dir_name
        sub_parms = parse_dm4_parms(dm4_file, sub_dir, dir_name)
        parm_dict.update(sub_parms)

    '''
    Loop over unnamed subdirectories
    '''
    for idir, sub_dir in enumerate(tag_dir.unnamed_subdirs):
        dir_name = '_'.join([base_name, 'SubDir_{:03d}'.format(idir)])
        if base_name == '':
            dir_name = 'Root'+dir_name
        sub_parms = parse_dm4_parms(dm4_file, sub_dir, dir_name)
        parm_dict.update(sub_parms)

    return parm_dict


def try_tag_to_string(tag_data):
    """
    Attempt to convert array of integers into a string

    Parameters
    ----------
    tag_data : array.array
        Array of 16-bit integers

    Returns
    -------
    tag_data : str
        Decoded string from the integer tag

    """
    if not isinstance(tag_data, array.array):
        return tag_data

    if tag_data.typecode == 'H':
        try:
            tag_data = str(tag_data.tostring().decode('utf-16'))
        except UnicodeDecodeError:
            pass
        except UnicodeEncodeError:
            pass
        except:
            raise

    return tag_data


def read_txt(image_path, header_lines=0, delimiter=None, *args, **kwargs):
    """

    Parameters
    ----------
    image_path : str
        Path to the image file
    header_lines : int
        Number of lines to skip as the header
    delimiter : str
        Separator between the columns of data
    args
    kwargs

    Returns
    -------
    image : numpy.ndarray
        Image array read from the plaintext file

    """
    image = np.loadtxt(image_path, *args,
                       skiprows=header_lines,
                       delimiter=delimiter, **kwargs)

    return image


def no_bin(image, *args, **kwargs):
    """
    Does absolutely nothing to the image.  Exists so that we can have
    a bin function to call whether we actually rebin the image or not.

    Parameters
    ----------
    image : ndarray
        Image
    args:
        Argument list
    kwargs:
        Keyword argument list

    Returns
    -------
    image : ndarray
        The input image
    """
    return image



"""
Created on Feb 9, 2016

@author: Chris Smith
"""

from __future__ import division, print_function, absolute_import, unicode_literals

import os

import numpy as np
from skimage.data import imread
from skimage.measure import block_reduce

from .df_utils.io_image import read_image, read_dm3, no_bin
from .translator import Translator
from .utils import generate_dummy_main_parms, build_ind_val_dsets
from ..hdf_utils import getH5DsetRefs, calc_chunks, link_as_main
from ..io_hdf5 import ioHDF5
from ..microdata import MicroDataGroup, MicroDataset


class PtychographyTranslator(Translator):
    """
    Translate Pytchography data from a set of images to an HDF5 file
    """
    def __init__(self, *args, **kwargs):
        super(PtychographyTranslator, self).__init__(*args, **kwargs)

        self.rebin = False
        self.bin_factor = 1
        self.hdf = None
        self.binning_func = no_bin
        self.bin_func = None
        self.image_ext = None

    def translate(self, h5_path, image_path, bin_factor=None, bin_func=np.mean, start_image=0, scan_size_x=None,
                  scan_size_y=None, image_type='.tif'):
        """
        Basic method that adds Ptychography data to existing hdf5 thisfile
        You must have already done the basic translation with BEodfTranslator
        
        Parameters
        ----------------
        h5_path : str
            Absolute path to where the HDF5 file should be located
        image_path : str
            Absolute path to folder holding the image files
        bin_factor : array_like of uint, optional
            Downsampling factor for each dimension.  Default is None.
        bin_func : callable, optional
            Function which will be called to calculate the return value
            of each block.  Function must implement an axis parameter,
            i.e. numpy.mean.  Ignored if bin_factor is None.  Default is
            numpy.mean.
        start_image : int, optional
            Integer denoting which image in the file path should be considered the starting
            point.  Default is 0, start with the first image on the list.
        scan_size_x : int, optional
            Number of Ronchigrams in the x direction.  Default is None, value will be determined
            from the number of images and `scan_size_y` if it is given.
        scan_size_y : int, optional
            Number of Ronchigrams in the y direction.  Default is None, value will be determined
            from the number of images and `scan_size_x` if it is given.
        image_type : str
            File extension of images to be read.  Default '.tif'

        Returns
        ----------
        h5_main : h5py.Dataset
            HDF5 Dataset object that contains the flattened images

        """
        self.image_ext = image_type

        # Open the hdf5 file and delete any contents
        try:
            hdf = ioHDF5(h5_path)
            hdf.clear()
        except:
            raise

        self.hdf = hdf

        # Get the list of all files with the .tif extension and the number of files in the list
        if image_type == '.dm3':
            file_list = [image_path]
            # image_path, _ = os.path.split(image_path)
            images, image_parms = read_dm3(image_path)
            usize = image_parms['SuperScan_Height']
            vsize = image_parms['SuperScan_Width']
            data_type = images.dtype
        else:
            file_list = self._parse_file_path(image_path, image_type)

            # Set up the basic parameters associated with this set of images
            (usize, vsize), data_type = self._getimagesize(os.path.join(image_path, file_list[0]))

        '''
        Check if a bin_factor is given.  Set up binning objects if it is.
        '''
        if bin_factor is not None:
            self.rebin = True
            if isinstance(bin_factor, int):
                self.bin_factor = (bin_factor, bin_factor)
            elif len(bin_factor) == 2:
                self.bin_factor = tuple(bin_factor)
            else:
                raise ValueError('Input parameter `bin_factor` must be a length 2 array_like or an integer.\n' +
                                 '{} was given.'.format(bin_factor))
            usize = int(usize / self.bin_factor[0])
            vsize = int(vsize / self.bin_factor[1])
            self.binning_func = block_reduce
            self.bin_func = bin_func

        if scan_size_x is None:
            scan_size_x = int(np.floor(np.sqrt(len(file_list))))
            scan_size_y = scan_size_x
        if scan_size_y is None:
            scan_size_y = int(np.floor(len(file_list) / scan_size_x))

        num_files = scan_size_x*scan_size_y

        h5_main, h5_mean_spec, h5_ronch = self._setupH5(usize, vsize, np.float32, scan_size_x, scan_size_y)

        self._read_data(file_list[start_image:start_image+num_files],
                        h5_main, h5_mean_spec, h5_ronch, image_path)

        return h5_main

    def _read_data(self, file_list, h5_main, h5_mean_spec, h5_ronch, image_path):
        """
        Iterates over the images in `file_list`, reading each image and downsampling if
        reqeusted, and writes the flattened image to file.  Also builds the Mean_Ronchigram
        and the Spectroscopic_Mean datasets at the same time.

        Parameters
        ----------
        file_list : list of str
            List of all files in `image_path` that will be read
        h5_main : h5py.Dataset
            Dataset which will hold the Ronchigrams
        h5_mean_spec : h5py.Dataset
            Dataset which will hold the Spectroscopic Mean
        h5_ronch : h5py.Dataset
            Dataset which will hold the Mean Ronchigram
        image_path : str
            Absolute file path to the directory which hold the images

        Returns
        -------
        None
        """

        mean_ronch = np.zeros(h5_ronch.shape, dtype=np.float32)

        num_files = len(file_list)

        for ifile, thisfile in enumerate(file_list):

            selected = (ifile + 1) % round(num_files / 16) == 0
            if selected:
                print('Processing file...{}% - reading: {}'.format(round(100 * ifile / num_files), thisfile))

            image, _ = read_image(os.path.join(image_path, thisfile), as_grey=True)
            image = self.binning_func(image, self.bin_factor, self.bin_func)
            image = image.flatten()
            h5_main[ifile, :] = image

            h5_mean_spec[ifile] = np.mean(image)

            mean_ronch += image

            self.hdf.flush()

        h5_ronch[:] = mean_ronch / num_files
        self.hdf.flush()

    # def downSampRoncVec(self, ronch_vec, binning_factor):
    #     """
    #     Downsample the image by taking the mean over nearby values
    #
    #     Parameters
    #     ----------
    #     ronch_vec : ndarray
    #         Image data
    #     binning_factor : int
    #         factor to reduce the size of the image by
    #
    #     Returns
    #     -------
    #     ronc_mat3_mean : ndarray
    #         Flattened downsampled image
    #     """
    #     ccd_pix = int(np.sqrt(ronch_vec.size))
    #     ronc_mat = ronch_vec.reshape(ccd_pix, ccd_pix)
    #     ronc_mat2 = ronc_mat.reshape(ccd_pix, ccd_pix / binning_factor, binning_factor)
    #     ronc_mat2_mean = ronc_mat2.mean(2)  # take the mean along the 3rd dimension
    #     ronc_mat3 = ronc_mat2_mean.reshape(ccd_pix / binning_factor, binning_factor, -1)
    #     ronc_mat3_mean = ronc_mat3.mean(1)
    #
    #     return ronc_mat3_mean.reshape(-1)

    @staticmethod
    def _parse_file_path(path, ftype='all'):
        """
        Returns a list of all files in the directory given by path
        
        Parameters
        ---------------
        path : string / unicode
            absolute path to directory containing files
        ftype : this file types to return in file_list. (optional. Default is all) 
        
        Returns
        ----------
        file_list : list of strings
            names of all files in directory located at path
        numfiles : unsigned int
            number of files in file_list
        """    
        
        # Get all files in directory
        file_list = os.listdir(path)
        
        # If no file type specified, return full list
        if ftype == 'all':
            return file_list

        # Remove files of type other than the request ftype from the list
        new_file_list = []
        for this_thing in file_list:
            # Make sure it's really a file
            if not os.path.isfile(os.path.join(path, this_thing)):
                continue

            split = os.path.splitext(this_thing)
            ext = split[1]
            if ext == ftype:
                new_file_list.append(os.path.join(path, this_thing))

        return new_file_list

    @staticmethod
    def _getimagesize(image):
        """
        Returns the x and y size of the image in pixels
        
        Parameters
        ------------
        image : string / unicode
            absolute path to the image file
        
        Returns
        -----------
        (size, tmp.dtype) : Tuple 
        
        size : unsigned integer
            x and y dimenstions of image
        dtype : data type
            Datatype of the image
        """
        tmp = imread(image)
        size = tmp.shape
        
        return size, tmp.dtype

    def _setupH5(self, usize, vsize, data_type, scan_size_x, scan_size_y):
        """
        Setup the HDF5 file in which to store the data including creating
        the Position and Spectroscopic datasets

        Parameters
        ----------
        usize : int
            Number of pixel columns in the images
        vsize : int
            Number of pixel rows in the images
        data_type : type
            Data type to save image as
        scan_size_x : int
            Number of images in the x dimension
        scan_size_y : int
            Number of images in the y dimension

        Returns
        -------
        h5_main : h5py.Dataset
            HDF5 Dataset that the images will be written into
        h5_mean_spec : h5py.Dataset
            HDF5 Dataset that the mean over all positions will be written
            into
        h5_ronch : h5py.Dataset
            HDF5 Dateset that the mean over all Spectroscopic steps will be
            written into
        """
        num_pixels = usize*vsize
        num_files = scan_size_x*scan_size_y

        root_parms = generate_dummy_main_parms()
        root_parms['data_type'] = 'PtychographyData'

        main_parms = {'num_images': num_files,
                      'image_size_u': usize,
                      'image_size_v': vsize,
                      'num_pixels': num_pixels,
                      'translator': 'Ptychography',
                      'scan_size_x': scan_size_x,
                      'scan_size_y': scan_size_y}
    # Create the hdf5 data Group
        root_grp = MicroDataGroup('/')
        root_grp.attrs = root_parms
        meas_grp = MicroDataGroup('Measurement_000')
        meas_grp.attrs = main_parms
        chan_grp = MicroDataGroup('Channel_000')
    # Get the Position and Spectroscopic Datasets
    #     ds_spec_ind, ds_spec_vals = self._buildspectroscopicdatasets(usize, vsize, num_pixels)
        ds_spec_ind, ds_spec_vals = build_ind_val_dsets((usize, vsize),
                                                        is_spectral=True,
                                                        labels=['U', 'V'],
                                                        units=['pixel', 'pixel'])
        ds_pos_ind, ds_pos_val = build_ind_val_dsets([scan_size_x, scan_size_y],
                                                     is_spectral=False,
                                                     labels=['X', 'Y'],
                                                     units=['pixel', 'pixel'])

        ds_chunking = calc_chunks([num_files, num_pixels],
                                  data_type(0).itemsize,
                                  unit_chunks=(1, num_pixels))

    # Allocate space for Main_Data and Pixel averaged Data
        ds_main_data = MicroDataset('Raw_Data', data=[], maxshape=(num_files, num_pixels),
                                    chunking=ds_chunking, dtype=data_type, compression='gzip')
        ds_mean_ronch_data = MicroDataset('Mean_Ronchigram',
                                          data=np.zeros(num_pixels, dtype=np.float32),
                                          dtype=np.float32)
        ds_mean_spec_data = MicroDataset('Spectroscopic_Mean',
                                         data=np.zeros(num_files, dtype=np.float32),
                                         dtype=np.float32)
    # Add datasets as children of Measurement_000 data group
        chan_grp.addChildren([ds_main_data, ds_spec_ind, ds_spec_vals, ds_pos_ind,
                              ds_pos_val, ds_mean_ronch_data, ds_mean_spec_data])
        meas_grp.addChildren([chan_grp])

        root_grp.addChildren([meas_grp])
        # print('Writing following tree to this file:')
        # root_grp.showTree()

        h5_refs = self.hdf.writeData(root_grp)
        h5_main = getH5DsetRefs(['Raw_Data'], h5_refs)[0]
        h5_ronch = getH5DsetRefs(['Mean_Ronchigram'], h5_refs)[0]
        h5_mean_spec = getH5DsetRefs(['Spectroscopic_Mean'], h5_refs)[0]
        aux_ds_names = ['Position_Indices',
                        'Position_Values',
                        'Spectroscopic_Indices',
                        'Spectroscopic_Values']

        link_as_main(h5_main, *getH5DsetRefs(aux_ds_names, h5_refs))

        self.hdf.flush()
        
        return h5_main, h5_mean_spec, h5_ronch

